from kubega.conftest import elastic_multiprocessing


@elastic_multiprocessing
def test_allreduce():
    import kubega.collective as collective
    import kubega.env as env

    collective.init("0.0.0.0")

    # allreduce with default reduce_fn (addition).
    result = collective.allreduce(env.replica_rank())
    assert result == sum(range(env.num_replicas()))

    # allreduce with custom reduce_fn (set union).
    result = collective.allreduce(
        {env.replica_rank()},
        reduce_fn=lambda a, b: a | b)
    assert result == set(range(env.num_replicas()))

    return [5, 0][env.num_restarts()]


@elastic_multiprocessing
def test_allreduce_async():
    import kubega.collective as collective
    import kubega.env as env

    collective.init("0.0.0.0")

    future_1 = collective.allreduce_async(1)
    future_2 = collective.allreduce_async(2)
    future_3 = collective.allreduce_async(3)
    assert future_2.result() == 2 * env.num_replicas()
    assert future_1.result() == 1 * env.num_replicas()
    assert future_3.result() == 3 * env.num_replicas()

    return [5, 0][env.num_restarts()]


@elastic_multiprocessing
def test_broadcast():
    import kubega.collective as collective
    import kubega.env as env

    collective.init("0.0.0.0")
    result = collective.broadcast(env.replica_rank())
    assert result == 0

    return [5, 0][env.num_restarts()]



"""
This module contains simple collective communications primitives which operate
on arbitrary python objects. It is meant to be general but *non-performant*.
Only use these primitives if you are synchronizing *small* objects which can be
efficiently pickled and operated on. For larger objects, use framework-specific
functions, such as those provided by `torch.distributed`.

The functions in this module should be invoked *in the same order* across all
replicas in the current job. Otherwise, their behavior is undefined, and you may
encounter unexpected bugs and errors.
"""
import logging
import pickle
import socket
import sys
import threading
import time
import traceback

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def default_reduce_fn(a, b):
    a += b
    return a


class Reducer:
    """
    Simple asynchronous allreduce operations on python objects.
    The replica with rank 0 is selected as the root pod, the left replicas are worker pods.

    Assume all invocations to allreduce, allreduce_async, and ``Future.result`` happen in the
    same order across all processes.
    """

    def __init__(self, rank, replicas, root_host, root_port):
        self._root_port = root_port
        self._result_map = {}
        self._next_key = 0
        self._rank = rank

        if rank == 0:
            # this is the root pod, start server on it
            self._reduce_fn_map = {}
            threading.Thread(target=self._run_server,
                             args=(self._root_port, replicas),
                             daemon=True).start()
        # Keep retrying connection, because
        # (1) the root pod might not have a registered domain name yet, and
        # (2) the root server socket might not be bound yet.
        exception_cnt = 0
        while True:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if exception_cnt > 25:
                logger.error("Could not connect to root after max retries, exiting...")
                break
            try:
                if self._root_port == 0:
                    # waiting for server to get a valid port in local mode
                    raise ConnectionRefusedError
                logger.info(f"rank {rank} of {replicas} connecting to "
                            f"{root_host} on port {self._root_port}")
                sock.connect((root_host, self._root_port))
            except ConnectionRefusedError:
                logger.warning("Could not connect to root, trying again...")
                exception_cnt += 1
                time.sleep(5)
            else:
                # if no error happens
                break
        self._sockfile = sock.makefile("rwb")
        pickle.dump(rank, self._sockfile)
        self._sockfile.flush()

    def _run_server(self, port, replicas):
        try:
            listener = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            listener.bind(("0.0.0.0", port))
            if port == 0:
                # local mode
                self._root_port = listener.getsockname()[1]
            listener.listen(replicas)
            # wait for connections from all clients
            logger.info(f"Master waiting for connections on {port}")
            clients = [None] * replicas
            while None in clients:
                client = listener.accept()[0].makefile("rwb")
                rank = pickle.load(client)
                assert clients[rank] is None
                clients[rank] = client

            # All connected. Now start the main loop.
            key = 0
            while True:
                # allreduce: reduce the result from root pod
                # reduce is implemented based on the received obj which are sent from clients
                for rank, client in enumerate(clients):
                    obj = pickle.load(client)
                    if rank == 0:
                        result = obj
                        reduce_fn = self._reduce_fn_map.pop(key)
                    else:
                        result = reduce_fn(result, obj)
                # Respond to clients in reverse order, with rank 0 last.
                # Prevent deadlocks where the rank 0 client gets unblocked
                # first and grab the GIL in a later operation, blocking this
                # server from responding to the remaining replicas.
                for client in reversed(clients):
                    pickle.dump((key, result), client)
                    client.flush()
                key += 1

        except Exception as e:
            traceback.print_exception(*sys.exc_info())
            logger.error(f"reducer server failed because of {e}")
            exit(1)

    def allreduce_async(self, obj, reduce_fn=default_reduce_fn):
        key = self._next_key
        self._next_key += 1
        try:
            self._reduce_fn_map[key] = reduce_fn
        except AttributeError:
            pass
        pickle.dump(obj, self._sockfile)
        self._sockfile.flush()
        return Future(self, key)

    def allreduce(self, obj, reduce_fn=default_reduce_fn):
        future = self.allreduce_async(obj, reduce_fn)
        return future.result()

    def broadcast(self, obj):
        """
        Use left projection to implement broadcast
        (Broadcast the reduced result from root pod to all work pods).
        """
        return self.allreduce(obj, lambda x, y: x)

    @property
    def result_map(self):
        return self._result_map

    @property
    def sockfile(self):
        return self._sockfile

    @property
    def rank(self):
        return self._rank


class Future:
    def __init__(self, reducer: Reducer, key):
        self._reducer = reducer
        self._key = key
        self._result = None

    def result(self):
        if self._result is None:
            while self._key not in self._reducer.result_map:
                try:
                    key, result = pickle.load(self._reducer.sockfile)
                    self._reducer.result_map[key] = result
                except Exception as e:
                    logger.error(f"reducer._rank = {self._reducer.rank} "
                                 f"is exiting unexpectedly because of {e}")
                    raise
            self._result = self._reducer.result_map.pop(self._key)
        return self._result



import functools
import time
import warnings
from abc import ABC
from typing import Optional

import numpy as np
import torch
import torch.cuda
import torch.distributed
from torch.autograd import Variable
from torch.nn.parallel import DistributedDataParallel

from kubega import checkpoint, env, utils
from kubega.torch._metrics import update_sync_time, update_grad_params, update_progress
from kubega.torch.data import current_dataloader
from kubega.torch.gradient_noise_scale import GradientNoiseScale, AdamGradientNoiseScale
from kubega.torch.scaling_rules import AdaScale, AdamScale, ScalingRuleBase


class _AdaptiveDataParallelState(checkpoint.State):
    def __init__(self, model, optimizer, lr_scheduler, mp_scaler, name="kubega-data-parallel"):
        super().__init__(name)
        self.model = model
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.mp_scaler = mp_scaler
        # TODO: Gain/Goodput should be tracked in the metrics module instead, not here.
        self.gain = 1.0
        # lr_factor summary
        self.lr_factor = 1.0

    def save(self, fileobj):
        state_dicts = [self.model.state_dict(), self.optimizer.state_dict()]

        if self.lr_scheduler is not None:
            state_dicts.append(self.lr_scheduler.state_dict())
        else:
            state_dicts.append(None)
        if self.mp_scaler is not None:
            state_dicts.append(self.mp_scaler.state_dict())
        else:
            state_dicts.append(None)

        torch.save((state_dicts, self.gain, self.lr_factor), fileobj)

    def load(self, fileobj):
        state_dicts, self.gain, self.lr_factor = torch.load(fileobj)
        self.model.load_state_dict(state_dicts[0])
        self.optimizer.load_state_dict(state_dicts[1])
        if state_dicts[2] is not None:
            self.lr_scheduler.load_state_dict(state_dicts[2])
        if state_dicts[3] is not None:
            self.mp_scaler.load_state_dict(state_dicts[3])


class AdaptiveDataParallel(DistributedDataParallel, ABC):
    """
    This class extends PyTorch DistributedDataParallel with support for
    adaptive batch sizes and checkpoint-restart elasticity. It automatically
    saves the given model, optimizer, and (optionally) LR scheduler whenever a
    checkpoint is triggered, and restores their states after restart. The
    optimizer is automatically patched with the chosen scaling rule.

    Arguments:
        model (torch.nn.Module): Model to be distributed.
        optimizer (torch.optim.Optimizer): Optimizer used to update the given
            model's parameters, will be patched using subclass of
        :class:`kubega.torch.scaling_rules.ScalingRuleBase`.
        scaling_rule (ScalingRuleBase): Scaling rule used to
            patch the given optimizer, default to AdaScale.
        lr_scheduler (torch.optim.lr_scheduler._LRScheduler): LR scheduler used
            to anneal the learning rate for the given optimizer.
        name (string): Unique name for each instance of this class, needed only
            if multiple instances exist.
    """

    def __init__(self, model, optimizer, lr_scheduler=None, mp_scaler=None,
                 scaling_rule: Optional[ScalingRuleBase] = None,
                 name="kubega-data-parallel", **kwargs):
        super().__init__(model, **kwargs)
        self._key = id(self)

        # Register backward hooks on model parameters. Depends on these hooks
        # being invoked before gradients are averaged. This is technically an
        # internal behavior of DistributedDataParallel, but seems to be abused
        # pretty widely so there should be little chance of it changing.
        # https://discuss.pytorch.org/t/59291
        for param in model.parameters():
            param.register_hook(functools.partial(self._backward_hook, param))

        # Setup for the scaling_rule, must be after registering backward hooks
        # because some of them need to register their own backward hooks.
        if not scaling_rule and (isinstance(optimizer, torch.optim.Adam) or
                                 isinstance(optimizer, torch.optim.AdamW)):
            self.scaling_rule = AdamScale()
        else:
            self.scaling_rule = scaling_rule or AdaScale()

        if isinstance(scaling_rule, AdamScale):
            self.gns = AdamGradientNoiseScale(self, optimizer, mp_scaler=mp_scaler)
        else:
            self.gns = GradientNoiseScale(self, optimizer, mp_scaler=mp_scaler)

        self.scaling_rule.initialize(self, optimizer, patch_optimizer=True)

        self._state = _AdaptiveDataParallelState(model, optimizer, lr_scheduler, mp_scaler, name)
        checkpoint.load_state(self._state)

        self._sync_start = None

    def forward(self, *args, **kwargs):
        # Do not do gradient synchronization during gradient accumulation.
        dataloader = current_dataloader()
        if dataloader is not None and dataloader.training:
            self.require_backward_grad_sync = dataloader.is_optim_step()
            accum_scale = (dataloader.current_local_bsz * env.num_replicas() / dataloader.batch_size)
            self.gns.set_accum_scale(accum_scale)
        return super().forward(*args, **kwargs)

    @utils.print_exc
    def _backward_hook(self, param, grad):
        # This method should be invoked once for each parameter during the
        # backward pass, before gradients are synchronized between replicas.
        if grad.device.type.startswith("cuda"):
            self._sync_start = torch.cuda.Event(enable_timing=True)
            self._sync_start.record()
        else:
            self._sync_start = time.time()
        self._final_callback_queued = False
        Variable._execution_engine.queue_callback(self._queue_callback)

    @utils.print_exc
    def _queue_callback(self):
        # This method should be invoked after the entire backward pass. We want
        # to make sure self._final_callback is invoked once, only after all
        # gradients have been synchronized between each replica. However, the
        # synchronization code in DistributedDataParallel is also done in a
        # callback, which might not yet be executed. Therefore, we enqueue
        # self._final_callback from this method, which should ensure it is
        # invoked after the gradient synchronization callback.
        if self._final_callback_queued:
            return
        self._final_callback_queued = True
        Variable._execution_engine.queue_callback(self._final_callback)

    @utils.print_exc
    def _final_callback(self):
        # This method should be invoked once for each backward pass, after
        # gradients have been synchronized between each replica.
        self._final_callback_queued = False
        # self._sync_start should mark the last time any local gradient
        # from this module was produced. We assume the duration until now was
        # primarily spent waiting for gradient synchronization.
        # TODO: Depends on the internal behavior of DistributedDataParallel,
        #       which might break with future versions of PyTorch. Any better
        #       and well-supported way to measure the synchronization time?
        if isinstance(self._sync_start, torch.cuda.Event):
            sync_end = torch.cuda.Event(enable_timing=True)
            sync_end.record()
            sync_end.synchronize()
            update_sync_time(self._sync_start.elapsed_time(sync_end) / 1e3)
        else:
            update_sync_time(time.time() - self._sync_start)

        dataloader = current_dataloader()
        if dataloader is None:
            # Don't allow backpropagation outside a dataloader loop, because
            # the batch size would be unknown.
            raise RuntimeError("backpropagation outside AdaptiveDataLoader")
        dataloader.train()

        scale = dataloader.current_batch_size / dataloader.batch_size
        self._state.gain = self.gns.gain(scale)
        self._state.lr_factor = \
            np.average(self.scaling_rule.scale_lr(scale))
        update_progress(self.gns.get_progress())
        if dataloader.max_batch_size and \
                dataloader.max_batch_size > dataloader.batch_size:
            update_grad_params(self._key, self.gns.sqr_avg(),
                               self.gns.var_avg())
        self._sync_start = None

    def zero_grad(self, *args, **kwargs):
        warnings.warn("zero_grad has no effect with AdaptiveDataParallel")

    @property
    def gain(self):  # TODO: should be tracked in the metrics module instead.
        """
        Current estimate of the AdaScale gain (r_t) value.
        """
        return self._state.gain

    def to_tensorboard(self, writer, global_step, tag_prefix=""):
        """
        Output some useful metrics to TensorBoard.

        Arguments:
            writer (torch.utils.tensorboard.SummaryWriter): ``SummaryWriter``
                object to output metrics to.
            global_step (int): Global step value to record.
            tag_prefix (str): Prefix added to each metric's tag.
        """
        if tag_prefix and not tag_prefix.endswith("/"):
            tag_prefix += "/"
        writer.add_scalar(tag_prefix + "Gradient_Norm_Sqr",
                          self.gns.sqr_avg(), global_step)
        writer.add_scalar(tag_prefix + "Gradient_Variance",
                          self.gns.var_avg(), global_step)
        writer.add_scalar(tag_prefix + "Gain",
                          self._state.gain, global_step)
        writer.add_scalar(tag_prefix + "Learning_Rate_Factor",
                          self._state.lr_factor, global_step)
        writer.add_scalar(tag_prefix + "Accum_Scale",
                          self.gns.accum_scale, global_step)
        if self.gns.accum_count > 0:
            writer.add_scalar(tag_prefix + "Accum_Count",
                              self.gns.accum_count, global_step)
        writer.add_scalar(tag_prefix + "Progress",
                          self.gns.get_progress(), global_step)



import pytest

from kubega.conftest import elastic_multiprocessing


@elastic_multiprocessing
def test_duplicate():
    from kubega.env import num_restarts
    from kubega.checkpoint import State
    s1 = State("state1")
    s2 = State("state2")
    with pytest.raises(ValueError):
        state_dup = State("state1")
    return [2, 0][num_restarts()]


@elastic_multiprocessing
def test_save_load():
    import pickle
    from kubega.checkpoint import State, save_all_states, load_state
    from kubega.env import replica_rank, num_restarts

    class TestState(State):
        def __init__(self, name):
            super().__init__(name)
            self.synced = False
            self.value = None

        def sync(self):
            self.synced = True

        def save(self, fileobj):
            assert replica_rank() == 0  # Should only be called from rank 0.
            pickle.dump(self.value, fileobj)

        def load(self, fileobj):
            # Should load the correct value.
            self.value = pickle.load(fileobj)

    s1 = TestState("state1")
    s2 = TestState("state2")

    if num_restarts() == 0:
        # Save all state.
        s1.value = 10
        s2.value = 20
        save_all_states()
        assert s1.synced and s2.synced
        return 2  # Restart with 2 replicas.
    elif num_restarts() == 1:
        # Load from checkpoints.
        load_state(s1)
        load_state(s2)
        assert s1.value == 10
        assert s2.value == 20
    else:
        assert False



from kubega.conftest import elastic_multiprocessing


@elastic_multiprocessing
def test_epoch():
    import kubega.checkpoint
    from kubega.env import num_restarts
    from kubega.torch.epoch import (remaining_epochs_until,
                                    current_epoch, finished_epochs)
    total_epochs = 10
    restart_epoch = 5
    assert current_epoch() is None
    if num_restarts() == 0:
        assert finished_epochs() == 0
        expected_epochs = list(range(restart_epoch + 1))
    elif num_restarts() == 1:
        assert finished_epochs() == restart_epoch
        expected_epochs = list(range(restart_epoch, total_epochs))
    else:
        assert False
    for idx, epoch in enumerate(remaining_epochs_until(10)):
        assert epoch == expected_epochs[idx]
        assert current_epoch() == epoch
        assert finished_epochs() == epoch
        if num_restarts() == 0 and epoch == restart_epoch:
            kubega.checkpoint.save_all_states()
            return 5  # Restart with 5 replicas.



"""
A wrapper of allreduce primitives.
"""
from kubega import env
from kubega.allreduce import Reducer, default_reduce_fn

_REDUCER = None


def init(master_addr=None, master_port=None, replica_rank=None, num_replicas=None):
    """
    Initialize this module, must be invoked before calling any other functions.
    This function will block until it has been invoked from all replicas.

    Arguments:
        master_addr: address of the replica with rank 0.
        master_port: free port of the replica with rank 0.
        replica_rank: rank of the current replica.
        num_replicas: total number of replicas.

    Raises:
        RuntimeError: If this module had already been initialized.
    """
    global _REDUCER
    if replica_rank is None:
        replica_rank = env.replica_rank()
    if num_replicas is None:
        num_replicas = env.num_replicas()

    if _REDUCER is not None:
        raise RuntimeError("{} is already initialized".format(__name__))

    if master_addr is None:
        master_addr = env.master_addr()
    if master_port is None:
        master_port = env.master_port()
    _REDUCER = Reducer(replica_rank, num_replicas, master_addr, master_port)


def teardown():
    """
    Teardown this module, will block until this function has been invoked from
    all replicas.

    Raises:
        RuntimeError: If this module has not been initialized.
    """
    raise NotImplementedError


def allreduce(value, reduce_fn=default_reduce_fn):
    """
    Reduces a value across all replicas in such a way that they all get the
    final result. Blocks until this function is invoked by all replicas.

    Arguments:
        value (object): The object which will be reduced together with all
            other replicas.
        reduce_fn (Function): A reduction function which two objects as
            arguments, and returns the resulting reduced object.

    Returns:
        object: Resulting value after being reduced across all replicas.

    Raises:
        RuntimeError: If this module has not been initialized.
    """
    if _REDUCER is None:
        raise RuntimeError("{} has not been initialized".format(__name__))
    return _REDUCER.allreduce(value, reduce_fn)


def allreduce_async(value, reduce_fn=default_reduce_fn):
    """
    Asynchronous version of the `allreduce` function. Does not block, instead
    returns a future which can be used to obtain the result later.

    Arguments:
        value (object): The object which will be reduced together with all
            other replicas.
        reduce_fn (Function): A reduction function which two objects as
            arguments, and returns the resulting reduced object.

    Returns:
        Future: Object from which the result can be obtained later.

    Raises:
        RuntimeError: If this module has not been initialized.
    """
    if _REDUCER is None:
        raise RuntimeError("{} has not been initialized".format(__name__))
    return _REDUCER.allreduce_async(value, reduce_fn)


def broadcast(value):
    """
    Broadcasts a value from the replica of rank 0 to all replicas. Blocks until
    this function is invoked by all replicas.

    Arguments:
        value (object): The object which will be broadcasted from replica 0.
            Ignored on all other replicas.

    Returns:
        object: The value broadcasted from replica 0.

    Raises:
        RuntimeError: If this module has not been initialized.
    """
    if _REDUCER is None:
        raise RuntimeError("{} has not been initialized".format(__name__))
    return _REDUCER.broadcast(value)



import os
import sys

if "darwin" in sys.platform.lower():
    # To avoid multiple runs of the model code
    # https://pythonspeed.com/articles/python-multiprocessing/
    import multiprocessing

    multiprocessing.set_start_method('fork')

import logging
import portpicker
import requests
import torch.distributed
import pkg_resources
import kubega.collective
import kubega.env
import semver
from kubega.torch.epoch import current_epoch, finished_epochs, remaining_epochs_until
from kubega.torch.data import current_dataloader, AdaptiveDataLoader, ElasticSampler
from kubega.torch.parallel import AdaptiveDataParallel
from kubega.torch.accumulator import Accumulator

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)

__all__ = [
    "init_process_group",
    "current_epoch",
    "finished_epochs",
    "remaining_epochs_until",
    "current_dataloader",
    "AdaptiveDataLoader",
    "ElasticSampler",
    "AdaptiveDataParallel",
    "Accumulator",
]


def version_check(version):
    if semver.VersionInfo.isvalid(version) and \
            version != "0.0.0":
        return True
    else:
        return False


def init_process_group(backend,
                       init_method=None,
                       world_size=None,
                       rank=None):
    """
    Initializes the default distributed process group and the kubeGA
    collectives module.

    Args:
        backend (str or Backend): The backend to use. Use "nccl" for multi-GPU
            training else "gloo".
        init_method (str, optional): URL specifying how to initialize the
                                     process group.
        world_size (int, optional): Number of processes participating in
                                    the job
        rank (int, optional): Rank of the current process (it should be a
                              number between 0 and ``world_size``-1).

    If init_method, world_size and rank is NOT provided, typically in the
    Kubernetes environment, kubeGA will try to infer them through environment
    variables KUBEGA_MASTER_ADDR, KUBEGA_NUM_REPLICAS and
    KUBEGA_REPLICA_RANK respectively.
    """
    if kubega.env.from_ray():
        from kubega_ray.kubega.utils import unique_nodes_pg
        assert init_method is not None
        assert world_size is not None
        assert rank is not None
        os.environ["kubega_NUM_NODES"] = str(unique_nodes_pg())
        os.environ["kubega_REPLICA_RANK"] = str(rank)
        os.environ["kubega_NUM_REPLICAS"] = str(world_size)

    url = kubega.env.supervisor_url()
    master_port = kubega.env.master_port()
    if rank is None:
        rank = kubega.env.replica_rank()

    if world_size is None:
        world_size = kubega.env.num_replicas()

    if init_method is not None:
        _, master_addr, master_port = init_method.split(":")
        master_addr = master_addr[2:]
        master_port = int(master_port)
    elif url:
        key = kubega.env.job_id()
        group = kubega.env.num_restarts()
        while True:
            response = requests.get(url=f"{url}/discover/{key}/{group}")
            if response.status_code != 408:  # Timeout.
                break
        response.raise_for_status()
        master_addr = response.json()[0]
        sched_version = kubega.env.kubega_sched_version()
        trainer_version = pkg_resources.get_distribution("kubega").version
        if version_check(sched_version) and version_check(trainer_version):
            trainer_ver_maj = semver.VersionInfo.parse(trainer_version).major
            sched_ver_maj = semver.VersionInfo.parse(sched_version).major
            if trainer_ver_maj != sched_ver_maj:
                raise Exception('kubega version {} is incompatible with'
                                'scheduler version {}'.format(trainer_version,
                                                              sched_version))
    else:
        master_addr = kubega.env.master_addr()

    # Initialize collective module.
    kubega.collective.init(master_addr, master_port, rank, world_size)

    # Initialize torch.distributed.
    torch_port = kubega.collective.broadcast(portpicker.pick_unused_port())
    init_method = "tcp://{}:{}?rank={}&world_size={}".format(
        master_addr, torch_port, rank, world_size)
    LOG.info("Initializing torch.distributed using %s", init_method)
    torch.distributed.init_process_group(backend, init_method)

    LOG.info("torch.distributed initialized")



import numpy as np
import torch
from torch.utils.data import Dataset

import kubega.torch as adl


class LRIterableDataset(Dataset):
    def __init__(self, size, true_values, noise):
        input_values = np.random.uniform(-5.0, 5.0, size)
        bias_input_values = np.stack([np.ones(size), input_values])
        target_values = np.dot(
            true_values, bias_input_values
        ) + np.random.normal(0.0, noise, size=(size,))
        self._values = list(zip(input_values, target_values))
        self._len = size

    def __getitem__(self, index):
        return self._values[index]

    def __len__(self):
        return self._len


def test_single_replica_parallel():
    adl.init_process_group("gloo")
    true_values = np.asarray([3.0, 4.0])
    dataset = LRIterableDataset(1000, true_values, 1.0)
    dataloader = adl.AdaptiveDataLoader(
        dataset, batch_size=32, shuffle=False, num_workers=1
    )
    model = torch.nn.Linear(1, 1, bias=True)
    params = [model.bias, model.weight]
    sgd = torch.optim.SGD([{"params": [param]} for param in params], lr=0.01)
    schedule = torch.optim.lr_scheduler.MultiStepLR(sgd, [50])
    model = adl.AdaptiveDataParallel(model, sgd, schedule)
    loss = torch.nn.MSELoss()
    for epoch in adl.remaining_epochs_until(100):
        for inputs, targets in dataloader:
            inputs = inputs.float()
            targets = targets.float()
            sgd.zero_grad()
            output = model(torch.reshape(inputs, (-1, 1)))
            targets = torch.reshape(targets, (-1, 1))
            loss_value = loss(output, targets)
            loss_value.backward()
            sgd.step()
        schedule.step()
    params = np.asarray([param.item() for param in params])
    assert np.all(np.isclose(params, true_values, atol=0.1)), (
        params,
        true_values,
    )



import random
from unittest.mock import Mock

import numpy as np
import pytest
import torch

from kubega.torch.gradient_noise_scale import GradientNoiseScale


def test_object():
    params = [
        torch.tensor([[1.0, -1.0], [2.0, 3.0]], requires_grad=True),
        torch.tensor([[2.0, 3.0]], requires_grad=True),
    ]
    sgd = torch.optim.SGD(params, lr=0.1)
    adp = Mock(require_backward_grad_sync=True)
    obj = GradientNoiseScale(adp, sgd, accum_scale=1.0, num_replicas=1)
    assert obj._accum_scale == 1.0
    obj._num_replicas = 8
    obj.set_accum_scale(3.0)
    assert obj.accum_scale == 3.0
    obj._num_replicas = 4
    obj.set_accum_scale(3.0)
    assert obj.accum_scale == 3.0
    assert np.isclose(obj.gain(2.0), 1.0)
    obj._state["var_avg"] = 3.0
    obj._state["norm_avg"] = 1.0
    assert np.isclose(obj.gain(3.0), 2.0)


ATOL = 0.01


def test_nan():
    def nan_objective(tensor):
        if random.random() > 0.5:
            target = float("Nan")
        else:
            target = 4.0
        return (tensor - target) ** 2

    params_t = torch.Tensor([1.0])
    params = torch.autograd.Variable(params_t, requires_grad=True)
    sgd = torch.optim.SGD([params], lr=0.1)
    adp = Mock(require_backward_grad_sync=True)
    gns = GradientNoiseScale(adp, sgd, accum_scale=1.0, num_replicas=1)
    adp.gns = gns
    for i in range(100):
        gns.reset_accumulation()
        loss = nan_objective(params)
        loss.backward()
        if np.all(np.isfinite(loss.detach().numpy())):
            sgd.step()
        if params.allclose(torch.tensor([4.0]), atol=ATOL):
            break
    else:
        pytest.fail(f"Did not converge: {params}")
    if not (
            np.all(np.isfinite(gns.sqr_avg()))
            and np.all(np.isfinite(gns.var_avg()))
    ):
        pytest.fail(
            f"non-finite adascale parameters:"
            f"{gns.sqr_avg()}, {gns.var_avg()}"
        )



import collections
import faulthandler
import signal
from multiprocessing import Process

import numpy as np
import portpicker

from kubega.allreduce import Reducer

root_host = "127.0.0.1"
DEFAULT_REDUCER_PORT = portpicker.pick_unused_port()


def init(rank, replicas):
    faulthandler.enable(all_threads=True)
    faulthandler.register(signal.SIGTERM, all_threads=True, chain=False)

    reducer = Reducer(rank, replicas, root_host, DEFAULT_REDUCER_PORT)
    if rank == 0:
        batch_size = 64
        x = {"foo": 1}
    else:
        x = {"bar": 1}
        batch_size = 0

    # Test 1: do a bunch of broadcasts
    # Because broadcast is performed by master, thus the broadcasted ``batch_size`` is 64, rather than 0
    batch_size = reducer.broadcast(batch_size)
    batch_size = reducer.broadcast(batch_size)
    batch_size = reducer.broadcast(batch_size)
    assert batch_size == 64

    # Test 2: do allreduce on Counter
    # the all reduced x is {"foo": 1, "bar": replicas - 1}
    x = reducer.allreduce(collections.Counter(x))
    assert x["foo"] == 1
    assert x["bar"] == replicas - 1

    # Test 3: do an async allreduce operation
    all_reduced_x = reducer.allreduce_async(np.asarray([1, 1, 1]))
    all_reduced_x = all_reduced_x.result()
    assert np.allclose(all_reduced_x, replicas * np.asarray([1, 1, 1]))

    # Test 4: simulate a training loop
    x = None
    for _ in range(10):
        if x:
            x = x.result()
            assert np.allclose(x, replicas * np.asarray([1, 1, 1]))
        x = reducer.allreduce_async(np.asarray([1, 1, 1]))


def test_allreduce():
    _replicas = 3
    processes = []
    for _rank in range(_replicas):
        p = Process(target=init, args=(_rank, _replicas), daemon=True)
        p.start()
        processes.append(p)

    for p in processes[1:]:
        p.join()

    processes[0].join()

    for p in processes:
        assert not p.exitcode



import functools
import math
import warnings
from types import MethodType

import numpy as np

from kubega.torch.data import current_dataloader

__all__ = ["ScalingRuleBase", "AdaScale", "AdamScale", "LinearScale", "SqrtScale", "LEGWScale"]


class ScalingRuleBase(object):
    """
    Base class for scaling rules that has the ability to track gradient noise
    scale calculations. Its subclasses can be used in combination with
    ``kubega.torch.parallel.AdaptiveDataParallel`` and ``torch.optim.SGD``.

    .. code-block:: python

        optim = torch.optim.SGD(model, lr=0.001)
        adascale = AdaScale()
        model = AdaptiveDataParallel(model, optim, adascale)

        for epoch in ...:
            for batch in ...:
                optim.zero_grad()
                loss = ...
                loss.backward()
                adascale.step()
    """

    def __init__(self):
        # instance of AdaptiveDataParallel, needs to be set before any of the
        # methods can be used
        self.adp = None
        self._optimizer = None
        self._orig_optimizer_step = None

    def scale_lr(self, scale):
        raise NotImplementedError

    def zero_grad(self, *args, **kwargs):
        if self.adp.gns.should_zero_grad:
            self.adp.gns.reset_accumulation(*args, **kwargs)
        else:
            warnings.warn("skipping zero_grad for accumulated gradient")

    def step(self, *args, **kwargs):
        """
        Run one optimizer step. Essentially just invokes
        ``optimizer.step(*args, **kwargs)`` with a scaled learning rate.

        Arguments:
            args: Positional arguments passed to ``optimizer.step``.
            kwargs: Keyword arguments passed to ``optimizer.step``.
        """
        if not self.adp:
            raise ValueError("AdaptiveDataParallel instance is not set!")
        if not self.adp.require_backward_grad_sync:
            return

        scale = self.adp.gns.accum_scale * self.adp.gns.accum_count
        initial_lr = [pg["lr"] for pg in self._optimizer.param_groups]
        scaled_lr = np.multiply(self.scale_lr(scale), initial_lr)
        for lr, pg in zip(scaled_lr, self._optimizer.param_groups):
            pg["lr"] = lr
        self._orig_optimizer_step(*args, **kwargs)
        for lr, pg in zip(initial_lr, self._optimizer.param_groups):
            pg["lr"] = lr

        self.adp.gns.set_progress(self.adp.gns.get_progress()
                                  + self.adp.gns.gain(scale))

    def _patch_optimizer(self):
        """
        Monkey-patch the optimizer's step function with
        :meth:`ScalingRuleBase.step`.
        """

        @functools.wraps(self._optimizer.step)
        def step_wrapper(optim, *args, **kwargs):
            return self.step(*args, **kwargs)

        @functools.wraps(self._optimizer.zero_grad)
        def zero_wrapper(optim, *args, **kwargs):
            return self.zero_grad(*args, **kwargs)

        self._optimizer.step = MethodType(step_wrapper, self._optimizer)
        self._optimizer.zero_grad = MethodType(zero_wrapper, self._optimizer)

    def initialize(self, adp, optimizer, patch_optimizer=False):
        self.adp = adp
        self._optimizer = optimizer
        self._orig_optimizer_step = optimizer.step
        if patch_optimizer:
            self._patch_optimizer()


class AdaScale(ScalingRuleBase):
    """
    Implements the AdaScale_ algorithm for scaling the learning rate for
    distributed and large batch size training.

    .. _AdaScale: https://proceedings.icml.cc/static/paper_files/icml/2020/4682-Supplemental.pdf
    """  # noqa: E501

    def scale_lr(self, scale):
        """Calculate factors to be applied to lr for each parameter group."""
        var = self.adp.gns.raw_var_avg
        sqr = self.adp.gns.raw_sqr_avg
        var = np.maximum(var, 1e-6)
        sqr = np.maximum(sqr, 0.0)
        return (var + sqr) / (var / scale + sqr)


class AdamScale(AdaScale):
    """
    Implements the variant of AdaScale_ that supports Adam, AdamW and RMSProp
    """

    def scale_lr(self, scale, power=0.5):
        return np.power(super().scale_lr(scale=scale), power)


class LinearScale(ScalingRuleBase):
    def scale_lr(self, scale):
        return scale


class SqrtScale(ScalingRuleBase):
    def scale_lr(self, scale):
        return math.sqrt(scale)


class LEGWScale(ScalingRuleBase):
    """
    Implements the LEGWScale algorithm for scaling the learning rate.

    Essentially, with LEGWScale, lr_factor is calculated based on
    training progress as follows:
    - when current_step < base_warmup_epoch * scale * steps_per_epoch:
      `lr_factor = sqrt(scale) * progress_ratio` where
      `progress_ratio = current_step /
                        (scale * base_warmup_epochs * steps_per_epoch)`
    - when current_step >= base_warmup_epoch * scale * steps_per_epoch:
      `lr_factor = sqrt(scale)`

    In order to adapt LEGWScale to AdaptDL, `progress_ratio` is
    calculated differently as:
    `progress / (scale * base_warmup_epochs * steps_per_epoch)` where
    `progress` is the effective steps trained based on AdaptDL's
    estimation.

    Arguments:
        base_warmup_epochs: Base warmup epochs
        data_size: total number of samples in the dataset

    .. _LEGWScale: https://arxiv.org/pdf/1901.08256.pdf
    """

    def __init__(self, base_warmup_epochs, data_size):
        super().__init__()
        self._base_warmup_epochs = base_warmup_epochs
        self._data_size = data_size

    def scale_lr(self, scale):
        dataloader = current_dataloader()
        # total training steps for warm up
        total_steps = self._base_warmup_epochs * scale * self._data_size / dataloader.batch_size
        max_lr_multiplier = math.sqrt(scale)
        # effective training steps taken
        progress = self.adp.gns.get_progress()
        if progress < total_steps:
            lr_factor = max_lr_multiplier * (progress / total_steps)
        else:
            lr_factor = max_lr_multiplier
        return lr_factor



"""
This module provides a function to post the optimized scheduling hints
(resource allocation details for a job) to the supervisor.
"""
import json
import logging
from collections import OrderedDict
from types import MappingProxyType

import requests

from kubega import env
from kubega.goodput import PerfParams

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)

PERF_PARAMS = MappingProxyType(OrderedDict({k: 0. for k in PerfParams._fields}))
SCHED_HINTS = MappingProxyType({'initBatchSize': 0,
                                'localBszBounds': None,  # [min, max]
                                'globalBatchSize': None,
                                'maxBatchSize': 0,
                                'maxProfiledReplicas': 0,
                                'gradientAccumulation': False,
                                'gradParams': None,
                                'perfParams': None})


def post_sched_hints(sched_hints, job_key):
    url = env.supervisor_url()
    if not url or url == "":
        return
    headers = {"Content-Type": "application/json"}
    try:
        for k in sched_hints:
            assert k in SCHED_HINTS

        response = requests.put(
            url=f"{url}/hints/{job_key}",
            data=json.dumps(sched_hints),
            headers=headers
        )
        if response.status_code != 200:
            LOG.warning(f"Received {response.status_code}")
    except Exception as e:
        LOG.warning(f"{e}")



import collections
import pickle
import time

import numpy as np

from kubega import checkpoint, env
from kubega.goodput import Goodput, fit_perf_params
from kubega.sched_hints import SCHED_HINTS, PERF_PARAMS, post_sched_hints

_METRICS_STATE = None
_GRAD_PARAM_DICT = {}
_PREV_REPORT_TIME = None


def _metrics_state():
    """
    Get metrics state from checkpoints.
    """
    global _METRICS_STATE
    if _METRICS_STATE is None:
        _METRICS_STATE = _MetricsState()
        checkpoint.load_state(_METRICS_STATE)
    return _METRICS_STATE


def _get_sched_hints():
    state = _metrics_state()
    if len(state.profile) == 0:
        return None
    _fit_perf_params()
    return _metrics_state()


def _post_sched_hints():
    assert env.replica_rank() == 0
    state = _metrics_state()
    sched_hints = SCHED_HINTS.copy()

    sched_hints["perfParams"] = {k: v for (k, v) in zip(PERF_PARAMS.keys(), state.perf_params)}
    sched_hints["maxBatchSize"] = state.max_batch_size
    sched_hints["localBszBounds"] = state.local_bsz_bounds
    sched_hints["initBatchSize"] = state.init_batch_size
    if state.grad_params:
        sched_hints["gradParams"] = {}
        sched_hints["gradParams"]["norm"] = state.grad_params[0]
        sched_hints["gradParams"]["var"] = state.grad_params[1]
    sched_hints["maxProfiledReplicas"] = max(key[1] for key in state.profile)
    sched_hints["gradientAccumulation"] = state.gradient_accumulation

    post_sched_hints(sched_hints, env.job_id())


def get_goodput():
    state = _metrics_state()
    if state.grad_params is None or state.perf_params is None:
        return None
    return Goodput(state.perf_params, state.grad_params, state.init_batch_size)


def profile_step_start(atomic_bsz):
    state = _metrics_state()

    state.atomic_bsz = atomic_bsz
    state.step_start = time.time()
    state.sync_time = 0.0


def update_batch_size(init_batch_size, max_batch_size, local_bsz_bounds, grad_accum):
    state = _metrics_state()

    state.init_batch_size = init_batch_size
    state.max_batch_size = max_batch_size
    state.local_bsz_bounds = local_bsz_bounds
    state.gradient_accumulation = grad_accum


def update_grad_params(edp_key, grad_norm_sqr, grad_variance):
    state = _metrics_state()

    global _GRAD_PARAM_DICT
    _GRAD_PARAM_DICT[edp_key] = np.asarray([grad_norm_sqr, grad_variance])
    grad_params = sum(_GRAD_PARAM_DICT.values())

    state.grad_params = (grad_params[0], grad_params[1])


def update_sync_time(sync_time):
    state = _metrics_state()
    state.sync_time += sync_time


def profile_step_commit(accum_step=False):
    global _PREV_REPORT_TIME
    state = _metrics_state()

    step_time = time.time() - state.step_start
    num_nodes = env.num_nodes()
    num_replicas = env.num_replicas()
    key = (num_nodes, num_replicas, state.atomic_bsz)
    if accum_step:
        state.profile[key]["accum_step_time"] += step_time
        state.profile[key]["accum_count"] += 1
    else:
        state.profile[key]["optim_step_time"] += step_time
        state.profile[key]["optim_sync_time"] += state.sync_time
        state.profile[key]["optim_count"] += 1
    del state.atomic_bsz
    del state.step_start
    del state.sync_time

    if not accum_step:
        if _PREV_REPORT_TIME is None:
            _PREV_REPORT_TIME = time.time()
        if env.replica_rank() == 0 and time.time() - _PREV_REPORT_TIME > 30:
            _fit_perf_params()
            _post_sched_hints()
            _PREV_REPORT_TIME = time.time()


def update_progress(progress):
    _metrics_state().progress = progress


def get_progress():
    return _metrics_state().progress


class _MetricsState(checkpoint.State):
    """
    The state to be saved into checkpoints in this design.
    """

    def __init__(self):
        super().__init__("kubega-metrics")
        self.profile = collections.defaultdict(collections.Counter)
        self.perf_params = None
        self.grad_params = None
        self.init_batch_size = None
        self.max_batch_size = None
        self.local_bsz_bounds = None
        self.gradient_accumulation = False
        self.progress = 0.0  # Progress in scale-invariant iterations.

    def save(self, fileobj):
        pickle.dump(self.profile, fileobj)
        pickle.dump(self.perf_params, fileobj)
        pickle.dump(self.grad_params, fileobj)
        pickle.dump(self.init_batch_size, fileobj)
        pickle.dump(self.max_batch_size, fileobj)
        pickle.dump(self.local_bsz_bounds, fileobj)
        pickle.dump(self.gradient_accumulation, fileobj)
        pickle.dump(self.progress, fileobj)

    def load(self, fileobj):
        self.profile = pickle.load(fileobj)
        self.perf_params = pickle.load(fileobj)
        self.grad_params = pickle.load(fileobj)
        self.init_batch_size = pickle.load(fileobj)
        self.max_batch_size = pickle.load(fileobj)
        self.local_bsz_bounds = pickle.load(fileobj)
        self.gradient_accumulation = pickle.load(fileobj)
        self.progress = pickle.load(fileobj)


def _fit_perf_params():
    """
    Load metrics from checkpoints and get the optimal resource allocation setting by fitting.
    """
    state = _metrics_state()
    profile = {k: v for k, v in state.profile.items() if v.get("optim_count")}
    num_nodes, num_replicas, atomic_bsz = (np.array(k) for k in zip(*profile.keys()))

    accum_step_time = np.array([v.get("accum_step_time", 0.0) for v in profile.values()])
    accum_count = np.array([v.get("accum_count", 0) for v in profile.values()])
    optim_step_time = np.array([v.get("optim_step_time", 0.0) for v in profile.values()])
    optim_sync_time = np.array([v.get("optim_sync_time", 0.0) for v in profile.values()])
    optim_count = np.array([v.get("optim_count", 0) for v in profile.values()])
    assert np.all(optim_count > 0)

    # Non-sync time during optimization steps should be approximately equal to
    # accumulation step time, combine those data points.
    assert np.all(optim_step_time >= optim_sync_time)
    accum_step_time += optim_step_time - optim_sync_time
    accum_count += optim_count
    accum_step_time /= accum_count
    optim_step_time /= optim_count
    state.perf_params = fit_perf_params(num_nodes, num_replicas, atomic_bsz,
                                        accum_step_time, optim_step_time)



"""
Implementation of gradient noise scale (GNS).
"""
import functools
import logging
import math

import numpy as np
import torch.distributed
import torch.optim
from torch.autograd import Variable

from kubega import utils

__all__ = ["GradientNoiseScale", "AdamGradientNoiseScale"]

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


def _calculate_preconditioner(param):
    return torch.ones_like(param, memory_format=torch.preserve_format)


def _average_groups(grads1, grads2):
    ret = []
    for group1, group2 in zip(grads1, grads2):
        ret.append([])
        for g1, g2 in zip(group1, group2):
            if g1 is None:
                ret[-1].append(g2)
            elif g2 is None:
                ret[-1].append(g1)
            else:
                ret[-1].append((g1 + g2) / 2)
    return ret


def _normsqr_groups(grads, pinvs):
    ret = []
    for group, pinv_group in zip(grads, pinvs):
        normsqr = [(g / pinv).pow(2).sum(dtype=torch.float64)
                   for g, pinv in zip(group, pinv_group) if g is not None]
        ret.append(sum(normsqr).item() if normsqr else 0.0)
    return np.array(ret)


class GradientNoiseScale:
    """
    This class tracks gradient related stats and takes care of gradient accumulation.
    """

    def __init__(self, adp, optimizer, mp_scaler=None, num_replicas=None, accum_scale=None):
        self._adp = adp
        self._optimizer = optimizer
        self._orig_optimizer_zero_grad = optimizer.zero_grad
        self._should_zero_grad = True
        self._mp_scaler = mp_scaler
        self._local_sqr = None
        self._accum_count = None
        self._num_replicas = (num_replicas if num_replicas is not None
                              else torch.distributed.get_world_size())
        self._accum_scale = accum_scale or self._num_replicas
        self._prev_grads = None

        self.reset_accumulation()

        self._optimizer.state.setdefault("gns", {
            "progress": 0.0,
            "prev_scale": 0.0,
            # Averages of n and v
            "sqr_avg": np.ones(len(optimizer.param_groups)),
            "var_avg": np.zeros(len(optimizer.param_groups)),
            # Whether estimates are biased (using difference estimation).
            "biased": False,
        })

        for idx, param_group in enumerate(self._optimizer.param_groups):
            for param in param_group["params"]:
                param.register_hook(functools.partial(self._backward_hook, idx, param))
        self._callback_queued = False
        self._smoothing = 0.999

    @property
    def _state(self):
        return self._optimizer.state["gns"]

    @property
    def should_zero_grad(self):
        return self._should_zero_grad

    @property
    def accum_scale(self):
        return self._accum_scale

    @property
    def accum_count(self):
        return self._accum_count

    def reset_accumulation(self):
        """reset accumulation calculations and gradients."""
        self._orig_optimizer_zero_grad()
        self._local_sqr = None
        self._accum_count = 0

    def set_accum_scale(self, accum_scale):
        if not np.isclose(self._accum_scale, accum_scale):
            self.reset_accumulation()
            self._accum_scale = accum_scale

    @property
    def raw_sqr_avg(self):
        view = self._state["sqr_avg"].view()
        view.flags.writeable = False
        return view

    def sqr_avg(self):
        """
        Current estimate of the squared l2-norm of the true gradient (sigma
        squared).

        Returns (float): Estimate of squared l2-norm.
        """
        return float(np.sum(np.maximum(self._state["sqr_avg"], 0.0)))

    @property
    def raw_var_avg(self):
        view = self._state["var_avg"].view()
        view.flags.writeable = False
        return view

    def var_avg(self):
        """
        Current estimate of the trace of the covariance of the true gradient
        (mu squared).

        Returns (float): Estimate of trace of the covariance.
        """
        return float(np.sum(np.maximum(self._state["var_avg"], 1e-6)))

    def get_progress(self):
        return self._state["progress"]

    def set_progress(self, progress):
        self._state["progress"] = progress

    def gain(self, scale):
        """
        Current estimate of the GradientNoiseScale gain ratio.

        Arguments:
            scale (float): The total scale to estimate the gain ratio for.

        Returns (float): Estimate of gain ratio.
        """
        var = self.var_avg()
        norm = self.sqr_avg()
        return (var + norm) / (var / scale + norm)

    def _update_avg(self, param_name, value, factor):
        biased = self._state.get(param_name + "_biased", 0.0)
        unbias = self._state.get(param_name + "_unbias", 0.0)
        biased = factor * biased + (1.0 - factor) * value
        unbias = factor * unbias + (1.0 - factor)
        self._state[param_name + "_biased"] = biased
        self._state[param_name + "_unbias"] = unbias
        self._state[param_name] = biased / unbias

    def _reset_avg(self, param_name):
        self._state.pop(param_name + "_biased", None)
        self._state.pop(param_name + "_unbias", None)

    @utils.print_exc
    def _backward_hook(self, idx, param, grad):
        # This method should be invoked once for each parameter during the
        # backward pass, before gradients are synchronized between replicas.
        if self._local_sqr is None:
            self._local_sqr = torch.zeros(len(self._optimizer.param_groups),
                                          device=grad.device,
                                          dtype=torch.float64)

        # Get the preconditioning matrix for the optimizer
        preconditioner = _calculate_preconditioner(param)

        # Update the local gradient square sum
        self._local_sqr[idx] += \
            (grad.detach() / preconditioner).pow(2).sum(dtype=torch.float64)

        if not self._callback_queued:
            Variable._execution_engine.queue_callback(self._queue_callback)
        self._callback_queued = True

    @utils.print_exc
    def _queue_callback(self):
        # This method should be invoked after the entire backward pass. We want
        # to make sure self._final_callback is invoked once, only after all
        # gradients have been synchronized between each replica. However, the
        # synchronization code in DistributedDataParallel is also done in a
        # callback, which might not yet be executed. Therefore, we enqueue
        # self._final_callback from this method, which should ensure it is
        # invoked after the gradient synchronization callback.
        self._callback_queued = False
        self._accum_count += 1
        if self._adp.require_backward_grad_sync:
            # Asynchronously sum the local squared-gradient statistics. The
            # actual gradient averaging should also be happening at the same
            # time, until self._final_callback is invoked.
            if self._num_replicas > 1:
                self._async_op = torch.distributed.all_reduce(self._local_sqr,
                                                              async_op=True)
            Variable._execution_engine.queue_callback(self._final_callback)
            self._should_zero_grad = True
        else:
            # Keep on accumulating gradients, should not zero grad.
            self._should_zero_grad = False

    @utils.print_exc
    def _final_callback(self):
        # This method should be invoked once the gradients have been
        # synchronized between all replicas and accumulation steps.
        if self._num_replicas > 1:
            self._async_op.wait()
        grads = []
        if self._mp_scaler is not None:
            mixed_precision_scale = self._mp_scaler.get_scale()
        else:
            mixed_precision_scale = 1.0
        for group in self._optimizer.param_groups:
            grads.append([])
            for param in group["params"]:
                if param.grad is None:
                    grads[-1].append(None)
                    continue
                param.grad.div_(self._accum_count)
                grads[-1].append(param.grad.detach().float() /
                                 mixed_precision_scale)
        preconditioner = self._get_preconditioner()

        # Note: mixed precision can result in nan/inf gradients,
        # which propagate into our norm and variance estimates.
        # Mixed precision autoscaling skips the skip where
        # there are nan/inf, so we also skip the update here
        grads_normsqr = _normsqr_groups(grads, preconditioner)
        if not np.all(np.isfinite(grads_normsqr)):
            LOG.warning(f"GradientNoiseScale detected invalid gradient! "
                        f"at scale {mixed_precision_scale}, Skipping step.")
            return
        count = self._num_replicas * self._accum_count
        scale = self._accum_scale * self._accum_count

        total_sqr = None
        local_sqr = None
        if count > 1:
            # Average local squared-norm samples.
            local_sqr = self._local_sqr.cpu().numpy() / count
            # Gradient is squared in local_sqr, so need to square the
            # mixed precision scale as well
            local_sqr = (local_sqr / mixed_precision_scale ** 2)
            total_sqr = grads_normsqr
            if self._state["biased"]:
                self._reset_avg("sqr_avg")
                self._reset_avg("var_avg")
            self._state["biased"] = False
            self._prev_grads = None
        else:
            # Single gradient datapoint, use difference estimation.
            if self._prev_grads is not None:
                local_sqr = (_normsqr_groups(self._prev_grads, preconditioner)
                             + grads_normsqr) / 2
                avg_grads = _average_groups(grads, self._prev_grads)
                total_sqr = _normsqr_groups(avg_grads, preconditioner)
                count = 2
                scale = 2 * self._accum_scale
            self._state["biased"] = True
            self._prev_grads = [[g.clone() if g is not None else None
                                 for g in group] for group in grads]
        if count > 1:
            grad_sqr = (count * total_sqr - local_sqr) / (count - 1)
            grad_var = (local_sqr - total_sqr) * scale / (count - 1)
            theta = self._smoothing ** scale
            self._update_avg('sqr_avg', grad_sqr, theta)
            self._update_avg('var_avg', grad_var, theta)

    def _get_preconditioner(self):
        out = []
        for idx, group in enumerate(self._optimizer.param_groups):
            pinvs = []
            for param in group["params"]:
                pinv = _calculate_preconditioner(param)
                pinvs.append(pinv)
            out.append(pinvs)
        return out


class AdamGradientNoiseScale(GradientNoiseScale):
    def __init__(self, adp, optimizer,
                 mp_scaler=None,
                 num_replicas=None,
                 accum_scale=None):

        self._adam_param_group = {'beta': [], 'eps': []}
        super().__init__(adp, optimizer, mp_scaler, num_replicas, accum_scale)
        for idx, param_group in enumerate(self._optimizer.param_groups):
            self._adam_param_group['beta'].append(param_group['betas'][1])
            self._adam_param_group['eps'].append(param_group['eps'])

    def _calculate_preconditioner(self, idx, param):
        state = self._optimizer.state[param]
        if state.get('step', 0) < 5:
            return torch.ones_like(param, memory_format=torch.preserve_format)

        exp_avg_sq = state["exp_avg_sq"].clone()  # not sure if clone is needed
        beta2 = self._adam_param_group['beta'][idx]
        eps = self._adam_param_group['eps'][idx]
        correction = 1 - beta2 ** state['step']
        pinv = (exp_avg_sq.sqrt() / math.sqrt(correction)).add_(eps)
        return pinv.to(param.device)

    def _reset_adam_state(self, step=0):
        for group in self._optimizer.param_groups:
            beta1, beta2 = group["betas"]
            for param in group["params"]:
                state = self._optimizer.state[param]
                if state.get("step", 0) > 0:
                    state["exp_avg"].mul_(
                        (1 - beta1 ** step) / (1 - beta1 ** state["step"]))
                    state["exp_avg_sq"].mul_(
                        (1 - beta2 ** step) / (1 - beta2 ** state["step"]))
                    state["step"] = step

    def _final_callback(self):
        scale = self._accum_scale * self._accum_count
        if not np.isclose(scale, self._state["prev_scale"]):
            # reset Adam states when scale is changed
            self._reset_adam_state()
            self._state["prev_scale"] = scale
        return super()._final_callback()



"""
Environment parameter retrieving.
"""
import os


def checkpoint_path():
    """
    Path to the directory used for saving and loading checkpoints. Determined
    by the environment variable ``KUBEGA_CHECKPOINT_PATH``, or ``None`` if
    unset. Setting this environment variable is required for checkpointing, and
    is automatically set in KUBEGA-scheduled clusters.

    Returns:
        str: checkpoint path or ``None``.
    """
    return os.getenv("KUBEGA_CHECKPOINT_PATH")


def share_path():
    """
    Path to a directory shared by all KUBEGA job replicas, which can be used
    by the application, e.g. for storing downloaded datasets or artifacts.
    Determined by the environment variable ``KUBEGA_SHARE_PATH``, or ``None``
    if unset. Automatically set in KUBEGA-scheduled clusters.

    Returns:
        str: shared directory path or ``None``.
    """
    return os.getenv("KUBEGA_SHARE_PATH")


def job_id():
    """
    A string which uniquely identifies the current job in an KUBEGA-scheduled
    cluster. ``None`` if running standalone.

    Returns:
        str: unique job identifier or ``None``.
    """
    return os.getenv("KUBEGA_JOB_ID")


def master_addr():
    """
    Network address of the rank 0 replica, required for distributed training.
    Determined by the environment variable ``KUBEGA_MASTER_ADDR``, or
    `0.0.0.0` if unset.

    In KUBEGA-scheduled clusters, this environment variable is unset. The rank
    0 replica is discovered dynamically by querying the supervisor
    (:func:`supervisor_url`).


    Returns:
        str: address of the rank 0 replica, or `0.0.0.0`.
    """
    return os.getenv("KUBEGA_MASTER_ADDR", "0.0.0.0")


def master_port():
    """
    Available port for the rank 0 replica, required for distributed training.
    Determined by the environment variable ``KUBEGA_MASTER_PORT``, or 0 if
    unset. Automatically set in KUBEGA-scheduled clusters.

    Returns:
        int: available port for the rank 0 replica, or 0.
    """
    return int(os.getenv("KUBEGA_MASTER_PORT", "0"))


def replica_rank():
    """
    Rank of the current replica, required for distributed training. Each
    replica is assigned a unique rank from 0 to K-1, where K is the total
    number of replicas. Determined by the environment variable
    ``KUBEGA_REPLICA_RANK``, or 0 if unset. Automatically set in
    KUBEGA-scheduled clusters.

    Returns:
        int: rank of the current replica, or 0.
    """
    return int(os.getenv("KUBEGA_REPLICA_RANK", "0"))


def num_nodes():
    """
    Number of unique nodes being used for the current job. For example, if
    there are 4 nodes, each running 2 replicas, then this function returns 4.
    Determined by the environment variable ``KUBEGA_NUM_NODES``, or is equal
    to :func:`num_replicas` if unset. Thus, this environment variable only
    needs to be set if some node runs multiple replicas. Automatically set in
    KUBEGA-scheduled clusters.

    Returns:
        int: number of unique nodes, or the value of :func:`num_replicas`.
    """
    return int(os.getenv("KUBEGA_NUM_NODES", num_replicas()))


def num_replicas():
    """
    Total number of replicas, required for distributed training. For example,
    if there are 4 nodes, each running 2 replicas, then this function returns
    8. Determined by the environment variable ``KUBEGA_NUM_REPLICAS``, or 1 if
    unset. Automatically set in KUBEGA-scheduled clusters.

    Returns:
        int: total number of replicas, or 1.
    """
    return int(os.getenv("KUBEGA_NUM_REPLICAS", "1"))


def num_restarts():
    """
    Number of times the current job was restarted. Determined by the
    environment variable ``KUBEGA_NUM_RESTARTS``, or 0 if unset. This value is
    mainly informational, and is automatically set in KUBEGA-scheduled
    clusters.

    Returns:
        int: number of restarts, or 0.
    """
    return int(os.getenv("KUBEGA_NUM_RESTARTS", "0"))


def kubega_sched_version():
    """
    A string which gives the KUBEGA version of scheduler. Determined
    by the environment variable ``KUBEGA_SCHED_VERSION`` or ``None``

    Returns:
        str: KUBEGA version of scheduler, or ``None``.
    """
    return os.environ.get("KUBEGA_SCHED_VERSION")


def supervisor_url():
    """
    URL of the supervisor in an KUBEGA-scheduled cluster. The address of the
    rank 0 replica is dynamically discovered via the supervisor, instead of via
    the ``KUBEGA_MASTER_ADDR`` environment variable.

    Returns:
        str: URL of the supervisor, or ``None``.
    """
    return os.getenv("KUBEGA_SUPERVISOR_URL")


def from_ray():
    """
    Returns True if the code is being called from Ray.
    """
    if os.getenv("KUBEGA_TUNE_TRIAL_SCHED", "False") == "True":
        import ray
        return ray.is_initialized()
    else:
        return False



from kubega.conftest import elastic_multiprocessing
from kubega.torch.accumulator import Accumulator


@elastic_multiprocessing
def test_accumulator_restarts():
    import kubega.checkpoint
    import kubega.collective
    from kubega.env import num_restarts, replica_rank
    kubega.collective.init("0.0.0.0")
    accum = Accumulator()

    if num_restarts() == 0:
        accum["a"] += 15  # Idempotent.
    assert "a" not in accum
    with accum.synchronized():
        assert "a" in accum
        assert accum["a"] == 15
    assert "a" not in accum
    if num_restarts() == 0:
        accum["a"] -= 5  # Idempotent.
        kubega.checkpoint.save_all_states()
        return 4  # Restart with 4 replicas.

    if num_restarts() == 1:  # Idempotent.
        accum.update({"a": replica_rank(), "b": replica_rank()})
    assert len(accum) == 0
    with accum.synchronized():
        assert len(accum) == 2
        assert accum["a"] == 16
        assert accum["b"] == 6
    assert len(accum) == 0
    if num_restarts() == 1:
        kubega.checkpoint.save_all_states()
        return 2  # Restart with 2 replicas.

    if num_restarts() == 2:  # Idempotent.
        accum -= {"b": 5, "c": 5}
    with accum.synchronized():
        assert accum["a"] == 16
        assert accum["b"] == -4
        assert accum["c"] == -10
        accum.clear()
    with accum.synchronized():
        assert not accum



"""
This module provides functionality to save and load arbitrary state as part of
checkpoint-restart elasticity. The `State` class can be subclassed to define
how to save/load any state to/from persistent storage, so it can be restored
after the current job restarts and resumed from where it left off.
"""
import logging
import os
import shutil

import kubega.env as env

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)

CKPT_DIR_PREFIX = "checkpoint-"

# FIXME: Keeping global state like this will result in memory leaks for
#  applications which do not restart too often.
_STATES2NAMES = {}
_NAMES2STATES = {}


class State:
    """
    This class implements an arbitrary piece of state which can be saved and
    loaded as part of a checkpoint, and synchronized across all replicas.
    Should be sub-classed to define custom save, load, and sync logic.
    """

    def __init__(self, name):
        """
        Initialize the state object with a unique identifier `name`, which is
        used to refer to the saved object in persistent storage. No two `State`
        objects may share the same `name`.

        Arguments:
            name (str): Unique name of this `State` object.

        Raises:
            ValueError: If a `State` object with the given name already exists.
        """
        if name in _NAMES2STATES:
            raise ValueError("State '{}' already exists".format(name))
        _NAMES2STATES[name] = self
        _STATES2NAMES[self] = name

    def save(self, fileobj):
        """
        This method should be overridden by subclasses to define how the state
        is saved. It is invoked by `save_all_states` and `save_state` to save the
        state into persistent storage.

        Arguments:
            fileobj (BinaryIO): A binary writable file object.
        """
        pass

    def load(self, fileobj):
        """
        This method should be overridden by subclasses to define how the state
        is loaded. It is invoked by `load_state` to load the state from persistent
        storage.

        Arguments:
            fileobj (BinaryIO): A binary readable file object.
        """
        pass

    def sync(self):
        """
        This method should be overridden by subclasses to define how the state
        is synchronized across replicas. This might be necessary to make sure
        the state is consistent before saving it to persistent storage. It is
        invoked by `save_state` before saving the state.
        """
        pass


def _get_tmp_ckpt_dir(checkpoint_path):
    if checkpoint_path is None:
        return None

    tmp_dir = os.path.join(checkpoint_path, "_checkpoint")
    os.makedirs(tmp_dir, exist_ok=True)
    return tmp_dir


def save_state(state, checkpoint_dir, sync=True):
    """
    Saves a `State` object to persistent storage. First invokes `State.sync` on
    all replicas if `sync` is `True` (default), and then invokes `State.save`
    on the replica of rank 0 only. Note that we save state to a temporary
    folder first. Then, it will be renamed to the formal checkpoint folder
    after all states are saved.

    Arguments:
        state (State): The `State` object to save to persistent storage.
        checkpoint_dir (string): The directory of checkpoint.
        sync (bool): Whether `State.sync` should be invoked.
    """
    if sync:
        state.sync()
    if env.replica_rank() == 0 and checkpoint_dir is not None:
        name = _STATES2NAMES[state]
        file = os.path.join(_get_tmp_ckpt_dir(checkpoint_dir), name)
        with open(file, "wb") as f:
            state.save(f)


def save_all_states():
    """
    Invokes `save_state` on all `State` objects for which `State.skip` is True.
    This function can be used to trigger a global checkpoint and save every
    `State` in the current job.
    """
    if env.from_ray():
        from ray.tune.trainable import TrainableUtil
        checkpoint_dir = TrainableUtil.make_checkpoint_dir("/tmp", index=None, override=True)
    else:
        checkpoint_dir = env.checkpoint_path()

    for state in _STATES2NAMES:
        save_state(state, checkpoint_dir)

    # Prevent corrupting original state files in case the process got killed
    # during state file writing.
    if env.replica_rank() == 0 and checkpoint_dir is not None:
        tmp_ckpt_dir = _get_tmp_ckpt_dir(checkpoint_dir)
        ckpt_dir = os.path.join(checkpoint_dir, f"{CKPT_DIR_PREFIX}{env.num_restarts()}")
        os.rename(tmp_ckpt_dir, ckpt_dir)  # atomic, rename(src, dst)
        for dir_name in os.listdir(checkpoint_dir):
            dir_path = os.path.join(checkpoint_dir, dir_name)
            if dir_name.startswith(CKPT_DIR_PREFIX) and dir_path != ckpt_dir:
                shutil.rmtree(dir_path)
        return checkpoint_dir


def load_state(state):
    """
    Load the given `State` object from persistent storage. If the object was
    previously saved, then `State.load()` will be invoked with a readable file
    object to load from.

    Arguments:
        state (State): `State` object to load from persistent storage.

    Returns:
        `True` if state was previously saved and `State.load` was invoked,
        `False` otherwise.
    """
    if env.from_ray():
        from ray.tune import session
        # TODO: Check this. get_checkpoint() --> loaded_checkpoint()?
        checkpoint_dir = session.get_session().loaded_checkpoint()
    else:
        checkpoint_dir = env.checkpoint_path()
    if checkpoint_dir is None:
        return False

    ckpt_dirs = os.listdir(checkpoint_dir)
    if not ckpt_dirs:
        LOG.info(f"No checkpoint found in {checkpoint_dir}.")
        return False

    latest_restart_id = 0
    for dir_name in ckpt_dirs:
        if dir_name.startswith(CKPT_DIR_PREFIX):
            restart_id = int(dir_name[len(CKPT_DIR_PREFIX):])
            latest_restart_id = max(latest_restart_id, restart_id)

    if latest_restart_id != env.num_restarts() - 1:
        LOG.warning("Cannot find checkpoint from the last restart. "
                    f"Loading checkpoint from restart {latest_restart_id}.")

    ckpt_dir = os.path.join(checkpoint_dir, f"{CKPT_DIR_PREFIX}{latest_restart_id}")
    name = _STATES2NAMES[state]
    state_file = os.path.join(ckpt_dir, name)
    if not os.path.isfile(state_file):
        LOG.warning(f"Cannot find state file {state_file}.")
        return False

    with open(state_file, "rb") as f:
        state.load(f)

    return True



"""
Create a temporary local environment which mimics a real DL job.
"""
import functools
import multiprocessing as mp
import os
import signal
import tempfile

import portpicker


def elastic_multiprocessing(func):
    """
    Decorator which runs a function inside a temporary local environment
    which mimics a real DL job. Runs replicas of the decorated function
    in their own processes, and sets up the shared environment, including
    environment variables and shared directories. The decorated function is
    always started with a single replica, but can optionally return an integer
    number of replicas to trigger a restart using that many replicas.

    ```python
    @elastic_multiprocessing
    def test_my_stuff():
        from KUBEGA.env import num_replicas, num_restarts
        if num_restarts() == 0:
            print(num_replicas)  # Outputs '1'.
            return 5  # Restart using 5 replicas.
        if num_restarts() == 1:
            print(num_replicas)  # Outputs '5'.
        return 0  # No more restarts, this line can be omitted.
    ```

    .. warning::
       The replica processes are forked from the current main process. This
       means that mutations to global variables in the main process prior to
       calling the decorated function may be observed by the child processes!
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        num_restarts = 0
        num_replicas = 1
        with tempfile.TemporaryDirectory() as tmpdir:
            while num_replicas:
                assert isinstance(num_replicas, int)
                master_port = portpicker.pick_unused_port()
                queue = mp.Queue()  # For passing return values back.

                def run(rank):  # Invoked in each child process.
                    os.environ["KUBEGA_CHECKPOINT_PATH"] = str(tmpdir)
                    os.environ["KUBEGA_JOB_ID"] = "tmpjob"
                    os.environ["KUBEGA_MASTER_PORT"] = str(master_port)
                    os.environ["KUBEGA_REPLICA_RANK"] = str(rank)
                    os.environ["KUBEGA_NUM_REPLICAS"] = str(num_replicas)
                    os.environ["KUBEGA_NUM_NODES"] = "1"
                    os.environ["KUBEGA_NUM_RESTARTS"] = str(num_restarts)
                    res = None
                    try:
                        res = func(*args, **kwargs)
                    finally:
                        queue.put((rank, res))

                # Start each replica in a separate child process.
                procs = [mp.Process(target=run, args=(rank,))
                         for rank in range(num_replicas)]
                for proc in procs:
                    proc.start()

                try:  # Wait for results from child processes.
                    for i in range(num_replicas):
                        rank, result = queue.get()
                        procs[rank].join()
                        assert procs[rank].exitcode == 0
                        if i == 0:  # All return values should be the same.
                            num_replicas = result
                        assert num_replicas == result
                finally:
                    # Clean up any remaining child processes.
                    for proc in procs:
                        if proc.is_alive():
                            os.kill(proc.pid, signal.SIGKILL)
                        proc.join()
                    # Clean up the queue.
                    queue.close()
                num_restarts += 1

    return wrapper



import itertools

import numpy as np
import pytest

from kubega import goodput
from kubega.goodput import Goodput, PerfParams, GradParams

RNG = np.random.RandomState(0)
PERF_PARAMS = [PerfParams(*RNG.gamma(2.0, 2.0, [7])) for _ in range(10)]
GRAD_PARAMS = [GradParams(*RNG.gamma(2.0, 2.0, [2])) for _ in range(10)]


def group_by_indices(*args):
    _, indices = np.unique(np.stack(args), axis=1, return_inverse=True)
    groups = {}
    for i, g in enumerate(indices):
        groups.setdefault(g, []).append(i)
    return list(groups.values())


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_evaluate(perf_params, grad_params):
    init_batch_size = 16
    goodput_fn = Goodput(perf_params, grad_params, init_batch_size)
    # Generate a range of different goodput function arguments.
    num_nodes = np.array([1, 2, 3, 4])
    num_replicas = np.array([1, 2, 4, 8])
    atomic_bsz = np.array([8, 12, 16, 20, 24])
    accum_steps = np.array([0, 1, 2, 3, 4])
    # Cartesian product.
    num_nodes, num_replicas, atomic_bsz, accum_steps = map(
        np.array,
        zip(*itertools.product(num_nodes, num_replicas, atomic_bsz, accum_steps)),
    )
    # Only keep valid arguments.
    valid = np.logical_and(
        num_nodes <= num_replicas,
        init_batch_size <= num_replicas * atomic_bsz * accum_steps,
    )
    num_nodes = num_nodes[valid]
    num_replicas = num_replicas[valid]
    atomic_bsz = atomic_bsz[valid]
    accum_steps = accum_steps[valid]
    # Evaluate goodput.
    goodput = goodput_fn(num_nodes, num_replicas, atomic_bsz, accum_steps)
    throughput = goodput_fn.throughput(
        num_nodes, num_replicas, atomic_bsz, accum_steps
    )
    efficiency = goodput_fn.efficiency(
        num_replicas * atomic_bsz * (accum_steps + 1)
    )
    # Check basic invariants.
    assert np.all(0 <= throughput)
    assert np.all(0 <= efficiency) and np.all(efficiency <= 1)
    assert np.allclose(goodput, throughput * efficiency)
    # Increasing batch size should decrease efficiency.
    batch_size = num_replicas * atomic_bsz * (accum_steps + 1)
    sort = np.argsort(batch_size)
    assert np.all(np.diff(efficiency[sort]) <= 0)
    # All else equal, increasing atomic_bsz should increase throughput.
    for indices in group_by_indices(num_nodes, num_replicas, accum_steps):
        sort = np.argsort(atomic_bsz[indices])
        assert np.all(np.diff(throughput[indices][sort]) >= 0)
        # Increasing throughput should experience diminishing returns.
        if len(indices) > 1:
            diffx = np.diff(atomic_bsz[indices][sort])
            diffy = np.diff(throughput[indices][sort])
            assert np.all(diffx[:-1] * diffy[1:] - diffx[1:] * diffy[:-1] <= 0)
    # All else equal, scalability is sublinear with respect to num_replicas.
    for indices in group_by_indices(num_nodes, atomic_bsz, accum_steps):
        scalability = throughput / num_replicas
        sort = np.argsort(num_replicas[indices])
        assert np.all(np.diff(scalability[indices][sort]) <= 0)


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_optimize_no_bounds(perf_params, grad_params):
    goodput_fn = Goodput(perf_params, grad_params, 128)
    goodput, bsz, steps = goodput_fn.optimize(1, 3)
    assert bsz == 128 // 3 + 1, "expected bsz = 43, got {}".format(bsz)
    assert isinstance(goodput, float)

    replicas = np.asarray([1, 2, 3, 4, 5])
    # single-node
    goodput, bsz, steps = goodput_fn.optimize(np.ones_like(replicas), replicas)
    assert bsz.shape == (5,)
    assert np.all(bsz == np.ceil(128 / replicas).astype(int))
    assert goodput.shape == (5,)
    assert bsz[0] == 128
    assert np.all(steps == 0)
    # multi-node
    goodput, bsz, steps = goodput_fn.optimize(replicas, replicas)
    assert bsz.shape == (5,)
    assert np.all(bsz == np.ceil(128 / replicas).astype(int))
    assert goodput.shape == (5,)
    assert bsz[0] == 128
    assert np.all(steps == 0)


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_optimize_local_bounds(perf_params, grad_params):
    fun = Goodput(perf_params, grad_params, 128)
    goodput, bsz, steps = fun.optimize(1, 1, atomic_bsz_range=(64, 256))
    assert bsz == 128, "expected bsz = 128, got {}".format(bsz)
    assert isinstance(goodput, float)

    replicas = np.asarray(range(1, 100))
    # single-node
    goodput, bsz, steps = fun.optimize(
        np.ones_like(replicas), replicas, atomic_bsz_range=(64, 256)
    )
    assert np.all(bsz >= np.ceil(128 / replicas).astype(int))
    assert np.all(np.logical_or(bsz >= 64, goodput == 0.0))
    assert np.all(bsz <= 256)
    assert np.all(bsz * replicas <= 100 * 128)
    assert bsz[0] == 128
    assert np.all(steps == 0)
    # multi-node
    goodput, bsz, steps = fun.optimize(
        replicas, replicas, atomic_bsz_range=(64, 256)
    )
    assert np.all(bsz >= np.ceil(128 / replicas).astype(int))
    assert np.all(np.logical_or(bsz >= 64, goodput == 0.0))
    assert np.all(bsz <= 256)
    assert np.all(bsz * replicas <= 100 * 128)
    assert bsz[0] == 128
    assert np.all(steps == 0)
    # closed bounds  w/o accumulation
    goodput, bsz, steps = fun.optimize(
        np.ones_like(replicas), replicas, atomic_bsz_range=(256, 256)
    )
    assert np.all(bsz >= np.ceil(128 / replicas).astype(int))
    assert np.all(np.logical_or(bsz == 256, goodput == 0.0))
    assert np.all(bsz == 256)
    assert bsz[0] == 256
    assert np.all(steps == 0)

    # closed bounds w/ accumulation
    goodput, bsz, steps = fun.optimize(
        np.ones_like(replicas),
        replicas,
        atomic_bsz_range=(256, 256),
        accumulation=True,
    )
    assert np.all(bsz >= np.ceil(128 / replicas).astype(int))
    assert np.all(np.logical_or(bsz == 256, goodput == 0.0))
    assert np.all(bsz == 256)
    assert bsz[0] == 256
    assert np.all(steps == 0)


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_optimize_max_bounds(perf_params, grad_params):
    fun = Goodput(perf_params, grad_params, 128)
    goodput, bsz, steps = fun.optimize(1, 1, max_batch_size=1280)
    assert bsz == 128, "expected bsz = 128, got {}".format(bsz)
    assert isinstance(goodput, float)

    replicas = np.asarray(range(1, 100))
    # single-node
    goodput, bsz, steps = fun.optimize(
        np.ones_like(replicas), replicas, max_batch_size=1280
    )
    assert np.all(bsz >= np.ceil(128 / replicas).astype(int))
    assert np.all(bsz * replicas <= 1280 + replicas)
    assert bsz[0] == 128
    assert np.all(steps == 0)
    # multi-node
    goodput, bsz, steps = fun.optimize(replicas, replicas, max_batch_size=1280)
    assert np.all(bsz >= np.ceil(128 / replicas).astype(int))
    assert np.all(bsz * replicas <= 1280 + replicas)
    assert bsz[0] == 128
    assert np.all(steps == 0)


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_optimize_all_bounds(perf_params, grad_params):
    fun = Goodput(perf_params, grad_params, 128)
    goodput, bsz, steps = fun.optimize(
        1, 1, max_batch_size=1280, atomic_bsz_range=(64, 256)
    )
    assert bsz == 128, "expected bsz = 128, got {}".format(bsz)
    assert isinstance(goodput, float)

    replicas = np.asarray(range(1, 20))
    # single-node
    goodput, bsz, steps = fun.optimize(
        np.ones_like(replicas),
        replicas,
        max_batch_size=1280,
        atomic_bsz_range=(64, 256),
    )
    assert np.all(
        np.logical_or(
            bsz >= np.ceil(128 / replicas).astype(int), goodput == 0.0
        )
    )
    assert np.all(np.logical_or(bsz >= 64, goodput == 0.0))
    assert np.all(bsz <= 256)
    assert np.all(
        np.logical_or(bsz * replicas <= 1280 + replicas, goodput == 0.0)
    )
    assert bsz[0] == 128
    assert np.all(steps == 0)
    # multi-node
    goodput, bsz, steps = fun.optimize(
        replicas, replicas, max_batch_size=1280, atomic_bsz_range=(64, 256)
    )
    assert np.all(
        np.logical_or(
            bsz >= np.ceil(128 / replicas).astype(int), goodput == 0.0
        )
    )
    assert np.all(np.logical_or(bsz >= 64, goodput == 0.0))
    assert np.all(bsz <= 256)
    assert np.all(
        np.logical_or(bsz * replicas <= 1280 + replicas, goodput == 0.0)
    )
    assert bsz[0] == 128
    assert np.all(steps == 0)
    # multi-node edge case
    replicas = 4
    goodput, bsz, steps = fun.optimize(
        4, 4, max_batch_size=1024, atomic_bsz_range=(128, 128)
    )
    assert goodput > 0.0
    assert bsz == 128
    assert steps == 0


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_optimize_accumulation(perf_params, grad_params):
    fun = Goodput(perf_params, grad_params, 128)
    goodput, bsz, steps = fun.optimize(
        1,
        1,
        max_batch_size=1280,
        atomic_bsz_range=(64, 256),
        accumulation=True,
    )
    assert isinstance(goodput, float)

    replicas = np.asarray(range(1, 20))
    # single-node
    goodput, bsz, steps = fun.optimize(
        np.ones_like(replicas),
        replicas,
        max_batch_size=1280,
        atomic_bsz_range=(64, 256),
        accumulation=True,
    )
    assert np.all(
        np.logical_or(
            bsz >= np.ceil(128 / replicas).astype(int), goodput == 0.0
        )
    )
    assert np.all(np.logical_or(bsz >= 64, goodput == 0.0))
    assert np.all(bsz <= 256)
    assert np.all(
        np.logical_or(
            bsz * replicas * (steps + 1) < 1280 + replicas * (steps + 1),
            goodput == 0.0,
        )
    )
    assert np.all(steps <= 15)
    assert np.all(steps >= 0)
    assert np.all(
        np.logical_or(replicas > 1, np.logical_or(bsz == 128, steps > 0))
    )
    # multi-node
    goodput, bsz, steps = fun.optimize(
        replicas,
        replicas,
        max_batch_size=1280,
        atomic_bsz_range=(64, 256),
        accumulation=True,
    )
    assert np.all(
        np.logical_or(
            bsz >= np.ceil(128 / replicas).astype(int), goodput == 0.0
        )
    )
    assert np.all(np.logical_or(bsz >= 64, goodput == 0.0))
    assert np.all(bsz <= 256)
    assert np.all(
        np.logical_or(
            bsz * replicas * (steps + 1) < 1280 + replicas * (steps + 1),
            goodput == 0.0,
        )
    )
    assert np.all(steps <= 15)
    assert np.all(steps >= 0)
    assert np.all(np.logical_or(np.multiply(steps, bsz) >= 256, steps == 0))


@pytest.mark.parametrize("perf_params", PERF_PARAMS)
@pytest.mark.parametrize("grad_params", GRAD_PARAMS)
def test_one_replica_accumulation(perf_params, grad_params):
    fun = Goodput(perf_params, grad_params, 128)

    replicas = np.asarray([1])
    max_batch_sizes = np.asarray(range(128, 128 * 20, 128))
    # single-node
    for max_batch_size in max_batch_sizes:
        goodput, bsz, steps = fun.optimize(
            np.ones_like(replicas),
            replicas,
            max_batch_size=1280,
            atomic_bsz_range=(64, 256),
            accumulation=True,
        )
        assert np.all(np.logical_or(bsz >= 64, goodput == 0.0))
        assert np.all(bsz <= 256)
        assert np.all(
            np.logical_or(bsz * (steps + 1) <= max_batch_size, goodput == 0.0)
        )
        assert np.all(
            np.logical_or(
                bsz >= np.ceil(128 / replicas).astype(int), goodput == 0.0
            )
        )
        assert np.all(np.logical_or(bsz * (steps + 1) != 128, steps == 0))


def test_fit_1():
    """
    Tests ``goodput.fit``'s ability to fit to data generated
    by its own model class with arbitrary parameters, without
    gradient accumulation. Serves as a sanity check
    that the ``goodput.model`` fitting works in the most
    optimistic case.
    """
    size = (1000,)
    nodes = np.random.randint(low=1, high=11, size=size)
    replicas = np.random.randint(low=1, high=nodes + 1, size=size)
    local_bsz = np.random.randint(32, 1024, size=size)
    params = goodput.PerfParams(0.1, 0.01, 0.5, 1.0, 1e-6, 1e-6, 1.2)

    accum_step_time = goodput._predict_accum_time(
        params, local_bsz
    ) + np.maximum(np.random.normal(0, 0.001, size=size), 0.0)

    network_time = goodput._predict_network_time(
        params, nodes, replicas
    ) + np.maximum(np.random.normal(0, 0.001, size=size), 0.0)

    gamma = params.gamma
    optim_step_time = (accum_step_time ** gamma + network_time ** gamma) ** (
            1 / gamma
    )

    result = goodput.fit_perf_params(
        nodes, replicas, local_bsz, accum_step_time, optim_step_time
    )

    loss_result = goodput._obj_fn(
        result, nodes, replicas, local_bsz, accum_step_time, optim_step_time
    )

    loss_true = goodput._obj_fn(
        params, nodes, replicas, local_bsz, accum_step_time, optim_step_time
    )

    assert (
            abs(loss_result - loss_true) < 0.1 * loss_true
            or loss_result < loss_true
    ), (
        "goodput.fit failed to fit model from data generated by",
        "goodput.PerfParams(0.1, 0.01, 0.5, 1.0, 1e-6, 1e-6, 1.2)",
        "parameters: {}".format(result),
    )


def test_fit_2():
    """
    Tests ``goodput.fit``'s ability to fit to data generated
    by its own model class with arbitrary parameters, without
    gradient accumulation. Serves as a sanity check
    that the ``goodput.model`` fitting works in the most
    optimistic case.
    """
    size = (1000,)
    nodes = np.random.randint(low=1, high=11, size=size)
    replicas = np.random.randint(low=1, high=nodes + 1, size=size)
    local_bsz = np.random.randint(32, 1024, size=size)
    params = goodput.PerfParams(0.1, 0.01, 0.5, 1.0, 1e-6, 1e-6, 1.2)

    accum_step_time = goodput._predict_accum_time(
        params, local_bsz
    ) + np.maximum(np.random.normal(0, 0.01, size=size), 0.0)

    network_time = goodput._predict_network_time(
        params, nodes, replicas
    ) + np.maximum(np.random.normal(0, 0.01, size=size), 0.0)

    gamma = params.gamma
    optim_step_time = (accum_step_time ** gamma + network_time ** gamma) ** (
            1 / gamma
    )

    result = goodput.fit_perf_params(
        nodes, replicas, local_bsz, accum_step_time, optim_step_time
    )

    loss_result = goodput._obj_fn(
        result, nodes, replicas, local_bsz, accum_step_time, optim_step_time
    )

    loss_true = goodput._obj_fn(
        params, nodes, replicas, local_bsz, accum_step_time, optim_step_time
    )

    assert (
            abs(loss_result - loss_true) < 0.1 * loss_true
            or loss_result < loss_true
    ), (
        "goodput.fit failed to fit model from data generated by",
        "goodput.PerfParams(0.1, 0.01, 0.5, 1.0, 1e-6, 1e-6, 1.2)",
        "parameters: {}".format(result),
    )



"""
This module provides tools for the top-level loop over epochs during training.
kubeGA expects the training program to be implemented as loop over several
epochs, each containing a series of loops over datasets (e.g. one loop over the
training set followed by one loop over the validation set). The program can be
interrupted between every iteration of any dataset loop, trigger a checkpoint
to be taken, and restarted using a different set of replicas.

**Due to checkpoint-restarts, parts of the training program may be executed
multiple times (e.g. once after each restart)!** To avoid incorrect execution,
ensure that your code is idempotent in the following locations:

1.  Immediately before any epoch loop (using :func:`remaining_epochs_until`).
2.  Immediately before any dataset loop (using
    :class:`kubega.torch.data.AdaptiveDataLoader`).

Your code may be non-idempotent in other locations.

.. code-block:: python

    ### IDEMPOTENT CODE ONLY ###
    for epoch in remaining_epochs_until(30):
        ### IDEMPOTENT CODE ONLY ###
        for batch in train_loader:
            # ... any code ...
        ### IDEMPOTENT CODE ONLY ###
        for batch in valid_loader:
            # ... any code ...
        # ... any code ...

    # ... any code ...

    ### END PROGRAM ###

For example, a common non-idempotent operation is learning-rate annealing:

.. code-block:: python

    for epoch in remaining_epochs_until(30):
        lr_scheduler.step()  # (A) WRONG!

        for batch in train_loader:
            # ...

        lr_scheduler.step()  # (B) WRONG!

        for batch in valid_loader:
            # ...

        lr_scheduler.step()  # (C) OK!

Location (A) will be executed again after any checkpoint-restart during either
the training or validation loop, resulting in the learning rate being annealed
several times in one epoch! Similarly, with location (B), if checkpoint-restart
happens during the validation loop.

Location (C) results in the correct behavior, because (1) an epoch will not be
repeated once it has finished, and (2) no checkpoint-restarts can occur between
the learning rate annealing and the end of the epoch.

.. _idempotent: https://stackoverflow.com/a/1077421
"""
import logging
import pickle

from kubega import checkpoint

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)

_EPOCH_STATE = None


def _epoch_state():
    global _EPOCH_STATE
    if _EPOCH_STATE is None:
        _EPOCH_STATE = _EpochState()
        checkpoint.load_state(_EPOCH_STATE)
    return _EPOCH_STATE


def current_epoch():
    """
    Get the current epoch while iterating with :func:`remaining_epochs_until`.

    Returns:
        int or None: The current epoch number if called from within a
        :func:`remaining_epochs_until` iteration, ``None`` otherwise.
    """
    return _epoch_state().current_epoch


def finished_epochs():
    """
    Get the number of epochs finished using :func:`remaining_epochs_until`.

    Returns:
        int: The number of finished epochs. Equal to :func:`current_epoch`
        if called from within a :func:`remaining_epochs_until` iteration.
    """
    return _epoch_state().finished_epochs


def remaining_epochs_until(epoch):
    """
    Iterate over epochs in a way that is consistent with checkpoint-restarts.
    For example:

    .. code-block:: python

        for epoch in remaining_epochs_until(30):
            print(current_epoch())  # Should print 0 through 29

        for epoch in remaining_epochs_until(60):
            print(current_epoch())  # Should print 30 through 59

    If a checkpoint-restart happens during an epoch, all previous epochs will
    be skipped after the program restarts.

    Arguments:
        epoch (int): The epoch number to end at (exclusively).

    Raises:
        RuntimeError: If invoked before a previous epoch loop has ended.
    """
    if current_epoch() is not None:
        raise RuntimeError("overlapping epoch loops detected")
    if finished_epochs() < epoch:
        LOG.info("starting at epoch %s", finished_epochs())
    else:
        LOG.info("skipping all epochs up to %s", epoch)
    while finished_epochs() < epoch:
        _epoch_state().current_epoch = finished_epochs()
        try:
            yield current_epoch()
        finally:
            # Try to catch any exits from epoch loop, including breaks and
            # Exceptions. See https://www.peterbe.com/plog/generatorexit.
            _epoch_state().finished_epochs += 1
            _epoch_state().current_epoch = None


class _EpochState(checkpoint.State):
    def __init__(self):
        super().__init__(".kubega-epoch")
        self.finished_epochs = 0
        self.current_epoch = None

    def save(self, fileobj):
        pickle.dump(self.finished_epochs, fileobj)

    def load(self, fileobj):
        self.finished_epochs = pickle.load(fileobj)



"""
The goodput metric and related operation.
"""
import collections
import numpy as np  # Reformat file with `Optimize imports` wil delete this! Thus, the error is caused.
import autograd
import scipy.optimize
import scipy.stats

"""
Parameters for a performance model which predicts the per-step time of
distributed SGD using allreduce. At a high level, models compute time and
network time separately, and combines them with some degree of overlap.
Compute time is modeled as a linear function of the local batch size.
Network time is modeled using different parameters depending on if the job
is inter-node (there exists a pair of replicas on different nodes), or
intra-node (all replicas are on the same node). For both cases, network time
is modeled as a constant term plus a retrogression term which increases
linearly with the total number of replicas.
"""
PerfParams = collections.namedtuple("PerfParams", [
    # T_compute ~ alpha_c + beta_c * local_bsz + (alpha_a + beta_a * local_bsz) * accumulation_steps
    "alpha_c",  # Constant term of compute time
    "beta_c",  # Multiplicative factor of compute time

    # If inter-node: T_network ~ alpha_n + beta_n * replicas
    "alpha_n",  # Constant term of inter-node network time
    "beta_n",  # Retrogression factor of inter-node network time

    # If intra-node: T_network ~ alpha_r + beta_r * replicas
    "alpha_r",  # Constant term of intra-node network time
    "beta_r",  # Retrogression factor of intra-node network time

    # T_step ~ (T_compute ^ gamma + T_network ^ gamma) ^ (1 / gamma)
    # Essentially T is a p-norm where p = gamma.
    # When p ~ 1 then T_step ~ T_compute + T_network, indicating no overlap between compute and network.
    # When p -> infinity then T_step = max(T_compute, T_network), indicating perfect overlap.
    #
    # We limit gamma to [1, 10] since 10 is close enough to approximate the max function for our purposes.
    "gamma",  # Models the degree of overlap between compute and network
])

GradParams = collections.namedtuple("GradParams", ["sqr", "var"])


class Goodput:
    def __init__(self, perf_params, grad_params, init_batch_size):
        self._perf_params = PerfParams(*perf_params)
        self._grad_params = GradParams(*grad_params)
        self._init_batch_size = init_batch_size

    def __call__(self, num_nodes, num_replicas, atomic_bsz, acc_steps):
        return self.evaluate(num_nodes, num_replicas, atomic_bsz, acc_steps)

    def evaluate(self, num_nodes, num_replicas, atomic_bsz, acc_steps):
        """
        Return the goodput value.
        """
        batch_size = num_replicas * atomic_bsz * (acc_steps + 1)
        assert np.all(self._init_batch_size <= batch_size)
        return self.throughput(num_nodes, num_replicas, atomic_bsz, acc_steps) * \
            self.efficiency(batch_size)

    def throughput(self, num_nodes, num_replicas, atomic_bsz, accum_steps):
        accum_time = _predict_accum_time(self._perf_params, atomic_bsz)
        network_time = _predict_network_time(self._perf_params,
                                             num_nodes, num_replicas)
        optim_time = np.exp(_predict_log_optim_time(self._perf_params,
                                                    accum_time, network_time))
        total_time = accum_steps * accum_time + optim_time
        batch_size = num_replicas * atomic_bsz * (accum_steps + 1)
        return batch_size / total_time

    def efficiency(self, batch_size):
        """
        TODO: Understand how this corresponds to formulas.
        """
        grad_sqr = self._grad_params.sqr
        grad_var = self._grad_params.var
        scale = batch_size / self._init_batch_size
        denom = grad_var / scale + grad_sqr
        gain = np.where(denom > 0, (grad_var + grad_sqr) / denom, 1.0)
        return gain / scale

    def optimize(self, num_nodes, num_replicas, max_batch_size=None,
                 atomic_bsz_range=None, accumulation=False):
        """
        Return the best goodput and corresponding settings (a, m, s).
        """
        assert np.all(np.less_equal(1, num_nodes))
        assert np.all(np.less_equal(num_nodes, num_replicas))
        if max_batch_size is None:
            max_batch_size = self._init_batch_size
        assert self._init_batch_size <= max_batch_size

        atomic_bsz_range = atomic_bsz_range or (None, None)
        min_atomic_bsz = atomic_bsz_range[0] or 1
        max_atomic_bsz = atomic_bsz_range[1] or max_batch_size

        # Remember what the output shape/format should be and flatten inputs.
        output_shape = np.broadcast(num_nodes, num_replicas).shape
        output_scalar = np.isscalar(num_nodes) or np.isscalar(num_replicas)
        num_nodes = np.broadcast_to(num_nodes, output_shape).flatten()
        num_replicas = np.broadcast_to(num_replicas, output_shape).flatten()

        # Samples 50 different total batch sizes in geometric space.
        min_batch_size = np.maximum(self._init_batch_size, min_atomic_bsz * num_replicas)
        batch_size = np.geomspace(min_batch_size, max_batch_size)
        local_bsz = batch_size / num_replicas
        eps = 1e-8  # Tolerance for floor/ceil operations.

        if accumulation:
            # If local_bsz size exceeds the max atomic batch size, split it
            # into a number of batches to form (atomic_bsz, accum_steps) such
            # that (atomic_bsz * (accum_steps + 1)) is close to local_bsz.
            #
            # If num_replicas == 1 and local_bsz > self._init_batch_size, then
            # set accum_steps to at least 1. This is because the gradient
            # statistics used for scaling up the learning rate are inaccurate
            # when there is only one atomic minibatch to estimate them from.
            accum_steps = np.ceil(local_bsz / max_atomic_bsz - eps) - 1
            accum_steps = np.where(
                np.logical_and(num_replicas == 1, local_bsz > self._init_batch_size + eps),
                np.maximum(accum_steps, 1),
                accum_steps
            ).astype(int)
            atomic_bsz = np.ceil(local_bsz / (accum_steps + 1) - eps).astype(int)

        else:
            accum_steps = np.zeros_like(local_bsz, dtype=np.int32)
            atomic_bsz = np.where(
                num_replicas == 1,
                self._init_batch_size,
                np.ceil(local_bsz - eps)
            ).astype(int)

        # Constrain the atomic_bsz before we evaluate the candidates
        atomic_bsz = np.maximum(min_atomic_bsz, atomic_bsz)
        atomic_bsz = np.minimum(max_atomic_bsz, atomic_bsz)

        # Evaluate the goodput of all candidate configurations.
        goodput = self.evaluate(num_nodes, num_replicas,
                                atomic_bsz, accum_steps)
        # Find the indices of the best configurations.
        indices = np.argmax(goodput, axis=0), np.arange(goodput.shape[1])

        # Restore the correct output shape and return results.
        goodput = goodput[indices].reshape(output_shape)
        atomic_bsz = atomic_bsz[indices].reshape(output_shape)
        accum_steps = accum_steps[indices].reshape(output_shape)
        if output_scalar:
            goodput = goodput.item()
            atomic_bsz = atomic_bsz.item()
            accum_steps = accum_steps.item()
        return goodput, atomic_bsz, accum_steps


def fit_perf_params(num_nodes, num_replicas, atomic_bsz, accum_step_time, optim_step_time):
    """
    Fit the performance model given accumulate time and optim time measurements
    for different configurations of num_nodes, num_replicas, and atomic_bsz.
    """
    # FIXME: We want to use the original numpy module for calls from the
    #  SpeedupFunction for performance reasons, but also need those functions to
    #  use autograd.numpy when we want to differentiate them. We patch the
    #  global np reference only for the code invoked from this function.
    global np  # Replace numpy from autograd.
    orig_np = np
    np = autograd.numpy

    num_nodes = np.array(num_nodes)
    num_replicas = np.array(num_replicas)
    accum_step_time = np.array(accum_step_time)
    optim_step_time = np.array(optim_step_time)

    # Set initial params to reasonable values.
    params = np.array([1e-1, 1e-2] * 3 + [1.0 + 1e-3])
    # Set lower/upper bounds for each parameter. Add a small slack to lower
    # bounds to avoid numerical instability issues.
    lower = [1e-8, 1e-8] * 3 + [1.0]
    upper = [np.inf, np.inf] * 3 + [10.0]
    if len(np.unique(atomic_bsz)) == 1:
        # Fix alpha_c if only observed a single atomic batch size.
        # This makes the speedup model optimistic with respect to
        # scaling up the batch size. This will assign equal weight
        # to the constant and multiplicative factors for accum time
        # if there is only a single datapoint (which is by far the
        # most likely case for this scenario)
        params[0] = upper[0] = lower[0] = np.mean(accum_step_time) / 2
    if not np.any(num_nodes > 1):
        # Fix alpha_n and beta_n if no multi-node observations.
        params[2] = upper[2] = lower[2]
        params[3] = upper[3] = lower[3]
    if not np.any(np.logical_and(num_nodes == 1, num_replicas > 1)):
        # Fix alpha_r and beta_r if no single-node/multi-replica observations.
        params[4] = upper[4] = lower[4]
        params[5] = upper[5] = lower[5]
    if not np.any(num_replicas > 2):
        # Fix beta_n and beta_r if no replicas > 2.
        params[3] = upper[3] = lower[3]
        params[5] = upper[5] = lower[5]
    bounds = scipy.optimize.Bounds(lower, upper, keep_feasible=True)
    args = (num_nodes, num_replicas, atomic_bsz,
            accum_step_time, optim_step_time)

    # FIXME: Need to handle optimization failures and propagate to the Trainer.
    grad_fn = autograd.grad(_obj_fn)
    result = scipy.optimize.minimize(_obj_fn, params, args=args,
                                     jac=grad_fn, bounds=bounds)
    params = result.x
    if not any(num_nodes > 1):
        # Enforce prior: alpha_n and beta_n are at least alpha_r and beta_r.
        params[2] = max(params[2], params[4] * 1.1)
        params[3] = max(params[3], params[5] * 1.1)

    np = orig_np  # Restore original numpy.

    return PerfParams(*params)


def _rmse(pred, true):
    return np.sqrt(((pred - true) ** 2).mean())


def _obj_fn(params, num_nodes, num_replicas, atomic_bsz, accum_step_time, optim_step_time):
    params = PerfParams(*params)
    pred_accum = _predict_accum_time(params, atomic_bsz)
    pred_network = _predict_network_time(params, num_nodes, num_replicas)
    pred_log_optim = _predict_log_optim_time(params, pred_accum, pred_network)

    # RMSLError of accum step time predictions.
    err1 = _rmse(np.log(pred_accum), np.log(accum_step_time))
    # RMSLError of optim step time predictions.
    err2 = _rmse(pred_log_optim, np.log(optim_step_time))

    # L2 regularization towards a smaller gamma, because it's easier to
    # optimize the alpha and beta parameters when gamma is smaller.
    reg1 = 1e-3 * (params.gamma - 1) ** 2

    # Penalize retrogression terms to prefer a more optimistic model.
    reg2 = 1e-2 * ((params.beta_n / params.alpha_n) ** 2 +
                   (params.beta_r / params.alpha_r) ** 2)

    return err1 + err2 + reg1 + reg2


def _predict_accum_time(params, atomic_bsz):
    params = PerfParams(*params)
    # Forward/backward passes should scale linearly with the batch size.
    return params.alpha_c + params.beta_c * atomic_bsz


def _predict_log_optim_time(params, acc_time, network_time):
    gamma = PerfParams(*params).gamma
    return np.log(acc_time ** gamma + network_time ** gamma) / gamma


def _predict_network_time(params, num_nodes, num_replicas):
    params = PerfParams(*params)
    # Select the most significant link between replicas, currently either
    # inter-node (nodes > 1) or intra-node (replicas > 1). Note that if
    # replicas == 1 then neither of these two conditions are matched.
    conditions = [num_nodes > 1, num_replicas > 1]
    # Bandwidth is bottle-necked by the most significant link, alpha models
    # the overhead of transferring data across that link.
    bottleneck = np.select(conditions, [params.alpha_n, params.alpha_r], 1e-8)
    # Assuming ring all-reduce, communication happens in a number of rounds
    # equal to the number of replicas. beta models the performance
    # retrogression from increasing the number of replicas beyond 2.
    retrogress = np.select(conditions, [params.beta_n, params.beta_r], 1e-8)
    retrogress = retrogress * np.maximum(num_replicas - 2, 1e-8)
    return bottleneck + retrogress



"""
This module provides the custom data sampler and dataloader with elastic batch size,
splitting over multiple replicas supported.
"""
import collections
import functools
import logging
import math
import pickle
import random
from contextlib import contextmanager

import numpy as np
import torch
from torch.utils.data import DataLoader, Sampler

from kubega import checkpoint, collective, env
from kubega._signal import get_exit_flag
from kubega.torch._metrics import (profile_step_start, profile_step_commit,
                                   update_batch_size, get_goodput, get_progress)
from kubega.torch.epoch import current_epoch

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


def current_dataloader():
    """
    Reference to the data loader currently being iterated.

    Returns (AdaptiveDataLoaderHelper): Current data loader.
    """
    return AdaptiveDataLoaderHelper.get_current_dataloader()


class ElasticSampler(Sampler):
    """
    A custom PyTorch Sampler which partitions data samples across multiple replicas,
    and supports deterministic continuing across checkpoint-restarts. Shuffling
    is deterministic for each epoch, and :meth:`ElasticSampler.set_epoch`
    should be invoked to obtain different orderings in different epochs.

    Arguments:
        dataset (torch.util.data.Dataset): The dataset to sample from.
        shuffle (bool): Whether the data samples should be shuffled.

    .. automethod:: __iter__
    .. automethod:: __len__
    """

    def __init__(self, dataset, shuffle=True):
        self.dataset = dataset
        self.shuffle = shuffle
        self.num_replicas = env.num_replicas()
        self.rank = env.replica_rank()
        self.epoch = 0
        self.index = 0

    def __iter__(self):
        """
        Iterate through the samples in the dataset, in the order defined for a
        set epoch, starting at a set index. Produces only the indices for the
        local replica.

        Returns: Iterator over data sample indices.
        """
        if self.shuffle:
            # Deterministically shuffle based on epoch.
            g = torch.Generator()
            g.manual_seed(hash((self.epoch, self.index // len(self.dataset))))
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = list(range(len(self.dataset)))

        base_index = self.index % len(self.dataset)

        # Subsample.
        local_indices = indices[base_index + self.rank::self.num_replicas]

        # Add extra samples to make it evenly divisible.
        if len(local_indices) < len(self):
            local_indices.append(indices[self.rank])
        assert len(local_indices) == len(self)
        return iter(local_indices)

    def __len__(self):
        """
        The total number of samples to be iterated through, starting at the set
        index, for the local replica.

        Returns (int): Number of samples.
        """
        base_index = self.index % len(self.dataset)
        return math.ceil((len(self.dataset) - base_index) / self.num_replicas)

    def set_epoch(self, epoch, index=0):
        """
        Set the epoch to derive samples from. Optional argument ``index`` can
        be specified to start sampling from a particular index, e.g. after a
        checkpoint-restart.

        Arguments:
            epoch (int): The epoch to sample from.
            index (int): The index to start sampling from.
        """
        self.epoch = epoch
        self.index = index


class _AdaptiveDataLoaderState(checkpoint.State):
    # Assume dataloaders are initialized in the same order in every replica.
    # Keep a map of epoch -> number of dataloaders initialized so far in that
    # epoch, and use that count to construct a unique name for the state.
    init_count = collections.Counter()

    def __init__(self):
        if current_dataloader() is not None:
            raise RuntimeError("dataloader may not be initialized during dataloader iteration")
        epoch = current_epoch()
        count = _AdaptiveDataLoaderState.init_count[epoch]
        super().__init__("kubega-dataloader-epoch{}-{}".format(epoch, count))
        _AdaptiveDataLoaderState.init_count[epoch] += 1

        self.current_index = 0  # Index within the current dataloader loop.
        self.end_index = 0  # End index of the current DataLoader loop.
        self.last_position = {}  # Epoch -> position of last completed loop.
        self.current_local_bsz = 0
        self.accumulation_steps = 0

    def save(self, fileobj):
        pickle.dump((self.current_index, self.end_index, self.last_position), fileobj)

    def load(self, fileobj):
        self.current_index, self.end_index, self.last_position = pickle.load(fileobj)


class AdaptiveDataLoaderHelper:
    """
    This class provides fine-grained control over adaptive training loops. It
    can be used for building more user-friendly custom data loaders, such as
    :class:`AdaptiveDataLoader`.

    Arguments:
        batch_size (int): The target total batch size across all replicas. The
            actual total batch size may be different due to rounding (each
            replica must have the same local batch size), or being scaled up
            using adaptive batch sizes.
    """

    # Epoch -> the number of dataloader loops completed so far in that epoch,
    # across all AdaptiveDataLoader objects.
    _position = collections.Counter()
    _training = None  # The AdaptiveDataLoader which loads training data.
    _current = None  # The AdaptiveDataLoader which is currently iterating.

    def __init__(self, batch_size=1):
        # Autoscale batch size fields.
        self._max_batch_size = None
        self._local_bsz_bounds = None

        # Create and load state.
        self._state = _AdaptiveDataLoaderState()
        checkpoint.load_state(self._state)

        self.batch_size = batch_size
        self.future_exit = None
        self._gradient_accumulation = False
        self._speedup_threshold = 1.05
        self._accum_count = 0

    @staticmethod
    def get_current_dataloader():
        return AdaptiveDataLoaderHelper._current

    @property
    def current_index(self):
        """
        The total number of data samples processed so far in the current loop.
        Includes the data processed by all replicas. ``None`` if this data
        loader is not currently being iterated.
        """
        if AdaptiveDataLoaderHelper._current is not self:
            return None
        return self._state.current_index

    @current_index.setter
    def current_index(self, index):
        if AdaptiveDataLoaderHelper._current is not self:
            return
        self._state.current_index = index

    @property
    def end_index(self):
        """
        (Optional) Can be used to track the end index of dataset across restarts.
        """
        return self._state.end_index

    @end_index.setter
    def end_index(self, index):
        """
        (Optional) Supports mutations of end_index.
        """
        self._state.end_index = index

    @property
    def max_batch_size(self):
        """
        The maximum total batch size allowed for adaptive batch size.
        ``None`` if adaptive batch size is disabled.
        """
        return self._max_batch_size

    @property
    def local_bsz_bounds(self):
        """
        The local batch size bounds on each replica. A pair of integers,
        (min_local_bsz, max_local_bsz).
        """
        return self._local_bsz_bounds

    @property
    def current_local_bsz(self):
        """
        The current logical local batch size used by the dataloader.
        The batch size returned by the dataloader may be smaller if
        gradient accumulation is used
        """
        return self._state.current_local_bsz

    @property
    def accumulation_steps(self):
        """
        The number of batches returned by the dataloader before a
        step is taken.
        """
        return self._state.accumulation_steps

    def is_accum_step(self):
        """
        Whether the current step's gradient will be accumulated.
        """
        return self._accum_count < self._state.accumulation_steps

    def is_optim_step(self):
        """
        Whether the optimizer step will be invoked in this step.
        """
        return not self.is_accum_step()

    def train(self):
        """
        Set this data loader to be the one used for training. Only one data
        loader may be used for training.
        """
        if AdaptiveDataLoaderHelper._training is None:
            AdaptiveDataLoaderHelper._training = self

        update_batch_size(self.batch_size, self.max_batch_size,
                          self.local_bsz_bounds, self._gradient_accumulation)

    def autoscale_batch_size(self, max_batch_size, local_bsz_bounds=None, gradient_accumulation=False):
        """
        Enables adaptive batch size. Should be invoked once after the data
        loader object is created.

        Arguments:
            max_batch_size (int): Maximum total batch size allowed.
            local_bsz_bounds (tuple): A pair of (min_local_bsz, max_local_bsz),
                the min and max local batch sizes allowed on each replica.
            gradient_accumulation (boolean): Gradient accumulation is adopted or not.

        Raises:
            ValueError: If any of the provided batch size bounds are invalid.
        """
        if not isinstance(max_batch_size, int) or \
                max_batch_size < self.batch_size:
            raise ValueError("invalid max_batch_size")

        if local_bsz_bounds is not None and (local_bsz_bounds[0] is not None and
                                             local_bsz_bounds[0] > self.batch_size or
                                             local_bsz_bounds[1] is not None and
                                             local_bsz_bounds[1] < self.batch_size):
            raise ValueError("invalid local_bsz_bounds")

        self._max_batch_size = max_batch_size
        self._local_bsz_bounds = local_bsz_bounds
        self._gradient_accumulation = gradient_accumulation
        self.train()

    def sync_local_bsz(self):
        goodput_fn = get_goodput()
        if self.max_batch_size is None or goodput_fn is None:
            # No autoscale batch size, just divide batch size evenly.
            self._state.current_local_bsz = math.ceil(
                self.batch_size / env.num_replicas())
            self._state.accumulation_steps = 0

        elif not self._state.current_local_bsz:
            # if init, use the batch size suggested
            _, atomic_bsz, accum_steps = goodput_fn.optimize(env.num_nodes(), env.num_replicas(),
                                                             max_batch_size=self._max_batch_size,
                                                             atomic_bsz_range=self._local_bsz_bounds,
                                                             accumulation=self._gradient_accumulation)
            self._state.current_local_bsz = atomic_bsz
            self._state.accumulation_steps = accum_steps

        else:
            # if not first time, we check against the relative speedup
            suggest_goodput, atomic_bsz, accum_steps = goodput_fn.optimize(env.num_nodes(), env.num_replicas(),
                                                                           max_batch_size=self._max_batch_size,
                                                                           atomic_bsz_range=self._local_bsz_bounds,
                                                                           accumulation=self._gradient_accumulation)

            # get current goodput
            current_goodput = goodput_fn(
                env.num_nodes(), env.num_replicas(),
                self.current_local_bsz, self.accumulation_steps)

            # use only if speedup is significant
            speedup = suggest_goodput / max(current_goodput, 1e-8)
            if speedup > self._speedup_threshold:
                self._state.current_local_bsz = atomic_bsz
                self._state.accumulation_steps = accum_steps

        self._state.current_local_bsz, self._state.accumulation_steps = collective.broadcast(
            (self._state.current_local_bsz, self._state.accumulation_steps)
        )
        return self.current_local_bsz

    @property
    def training(self):
        return self is AdaptiveDataLoaderHelper._training

    @contextmanager
    def profile(self, commit):
        """
        Every iteration of every epoch should be profiled under this context.
        Note that, custom DataLoader writers should make sure that it gets
        called equal number of times on each replica.

        Arguments:
            commit (bool): Whether to commit the profiled results.
        """
        # Synchronize the exit signal so all replicas exit after
        # the same iteration. Do this asynchronously to prevent
        # unnecessary blocking on the network.
        if self.future_exit is not None and self.future_exit.result():
            checkpoint.save_all_states()
            exit(143)  # Standard exit code response to SIGTERM.
        self.future_exit = collective.allreduce_async(get_exit_flag(), lambda a, b: a or b)
        profile_step_start(self.current_local_bsz)

        yield

        if commit:
            profile_step_commit(self.is_accum_step())
        self._accum_count = 0 if self.is_optim_step() else self._accum_count + 1

    @contextmanager
    def context(self):
        """
        All iterators should be iterated under this context. It ensures
        proper cleanup of elastic context at the end of each epoch.
        """
        epoch = current_epoch()
        try:
            if AdaptiveDataLoaderHelper._current is not None:
                raise RuntimeError("overlapping dataloader \
                                    iterations detected")
            AdaptiveDataLoaderHelper._current = self

            yield
        finally:
            self._state.current_index = 0
            self._state.end_index = 0
            self._state.last_position[epoch] = self._position[epoch]
            self._position[epoch] += 1
            AdaptiveDataLoaderHelper._current = None

    @property
    def current_batch_size(self):
        return self.current_local_bsz * (self.accumulation_steps + 1) * env.num_replicas()

    def skipdone(self):
        """
        Should be called just after entering the `_elastic` context to make
        sure that the dataloader loop is not replayed if it has already finished
        before a restart.
        """

        epoch = current_epoch()
        position = self._position[epoch]
        if position <= self._state.last_position.get(epoch, -1):
            # Already completed the dataloader loop at the current position,
            # skip this loop and keep replaying the application code.
            LOG.info("skipping %s loop at position %s in epoch %s",
                     self.__class__.__name__, position, epoch)
            self._position[epoch] += 1
            return True
        else:
            return False

    def to_tensorboard(self, writer, global_step, tag_prefix=""):
        """
        Output some useful metrics to TensorBoard.

        Arguments:
            writer (torch.utils.tensorboard.SummaryWriter): ``SummaryWriter``
                object to output metrics to.
            global_step (int): Global step value to record.
            tag_prefix (str): Prefix added to each metric's tag.
        """
        if tag_prefix and not tag_prefix.endswith("/"):
            tag_prefix += "/"

        writer.add_scalar(tag_prefix + "Total_Batch_Size",
                          self.current_batch_size, global_step)

        writer.add_scalar(tag_prefix + "Local_Batch_Size",
                          self.current_local_bsz, global_step)

        writer.add_scalar(tag_prefix + "Accumulation_Steps",
                          self.accumulation_steps, global_step)


class AdaptiveDataLoaderMixin:
    """
    This class provides elastic functionality to any custom DataLoader which
    inherits it. It defines a member _elastic of type
    :class:`AdaptiveDataLoaderHelper` which has useful methods and members to
    implement restart-safe, elastic DataLoaders. It also exposes public methods
    which can be used inside training loops directly from
    :class:`AdaptiveDataLoader`.

    This is a wrapper of :class:`AdaptiveDataLoaderHelper`.
    """

    def __init__(self, batch_size):
        self._elastic = AdaptiveDataLoaderHelper(batch_size)

    def autoscale_batch_size(self, max_batch_size,
                             local_bsz_bounds=None,
                             gradient_accumulation=False):
        self._elastic.autoscale_batch_size(max_batch_size, local_bsz_bounds, gradient_accumulation)

    @property
    def current_local_bsz(self):
        if current_dataloader() is not self._elastic:
            return None
        return self._elastic.current_local_bsz

    @property
    def accumulation_steps(self):
        """
        The number of batches returned by the dataloader before a
        step is taken.
        """
        return self._elastic.accumulation_steps

    @property
    def training(self):
        return self._elastic.training

    @property
    def current_batch_size(self):
        if current_dataloader() is not self._elastic:
            return None
        return self._elastic.current_batch_size

    def to_tensorboard(self, writer, global_step, tag_prefix=""):
        self._elastic.to_tensorboard(writer, global_step, tag_prefix)

    to_tensorboard.__doc__ = AdaptiveDataLoaderHelper.to_tensorboard.__doc__


class AdaptiveDataLoader(DataLoader, AdaptiveDataLoaderMixin):
    """
    This class is a PyTorch DataLoader that also supports adaptive batch sizes
    and checkpoint-restart elasticity. Applications can typically use objects
    of this class as direct replacements for PyTorch DataLoaders. However, some
    notable differences are:

    1.  The ``batch_size`` argument defines the target total batch size across
        all replicas, rather than the local batch size on each replica.
    2.  Custom ``sampler`` and ``batch_sampler`` are not supported.
    3.  Iterating through the dataloader is only allowed from within an epoch
        loop (see :mod:`kubega.torch.epoch`), and only one dataloader loop is
        allowed at any given time.

    Arguments:
        dataset (torch.util.data.Dataset): Dataset from which to load the data.
        batch_size (int): The target total batch size across all replicas. The
            actual total batch size may be different due to rounding (each
            replica must have the same local batch size), or being scaled up
            using adaptive batch sizes.
        shuffle (bool): Whether the data is reshuffled at every epoch.
        **kwargs: Keyword arguments passed to ``torch.util.data.Dataloader``.

    Raises:
        ValueError: If ``sampler`` or ``batch_sampler`` are not ``None``.

    .. automethod:: __iter__
    """

    def __init__(self, dataset, batch_size=1, shuffle=False, **kwargs):
        if kwargs.get("batch_sampler") is not None \
                or kwargs.get("sampler") is not None:
            raise ValueError("AdaptiveDataLoader does not support custom 'sampler' or 'batch_sampler'")

        # Custom sampler is incompatible with shuffle=True, so we always set
        # shuffle=False in __init__ and let our own sampler do the shuffling.
        kwargs["sampler"] = ElasticSampler(dataset, shuffle=shuffle)
        kwargs["worker_init_fn"] = _worker_init_wrapper(kwargs.get("worker_init_fn"), kwargs.get("num_workers"))

        super().__init__(dataset, batch_size, shuffle=False, **kwargs)
        AdaptiveDataLoaderMixin.__init__(self, batch_size)

    def __iter__(self):
        """
        Iterate over batches of data. When adaptive batch size is disabled,
        stops after the entire dataset has been processed once in total by all
        replicas. This means if there are K replicas, then this method will
        iterate over ~1/K of the dataset. When adaptive batch size is enabled,
        stops after making enough statistical progress roughly equivalent to
        one pass over the dataset with non-adaptive batch size. In this case,
        the dataset may be processed more than once.

        A checkpoint-restart may be triggered in-between each batch. In this
        case, the current iteration state will be saved and restored after the
        restart, and continue where it left off.
        """
        epoch = current_epoch()
        num_replicas = env.num_replicas()
        with self._elastic.context():
            if self._elastic.skipdone():
                return
            done = False
            while not done:
                self.sampler.set_epoch(epoch, index=self._elastic.current_index)
                self.batch_sampler.batch_size = self._elastic.sync_local_bsz()
                for idx, batch in enumerate(super().__iter__()):
                    with self._elastic.profile(self.training and idx >= 1):
                        yield batch
                        # Increment by the number of data samples processed
                        self._elastic.current_index += \
                            num_replicas * self.batch_sampler.batch_size
                        if self._elastic.max_batch_size is not None and \
                                get_progress() >= len(self.dataset) * (epoch + 1) / self.batch_size:
                            done = True
                            break
                if self._elastic.max_batch_size is None:
                    done = True
                self._elastic.current_index -= \
                    self._elastic.current_index % -len(self.dataset)


def _worker_init_wrapper(worker_init_fn, num_workers):
    # Set globally-unique python and numpy seeds for each worker
    @functools.wraps(worker_init_fn)
    def wrapper(worker_id):
        nonlocal num_workers
        num_workers = num_workers or 1
        # https://pytorch.org/docs/master/data.html#randomness-in-multi-process-data-loading.
        seed = torch.initial_seed() + env.replica_rank() * num_workers
        torch.manual_seed(seed)
        np.random.seed(seed % 2 ** 32)
        random.seed(seed)
        if worker_init_fn is not None:
            return worker_init_fn(worker_id)

    return wrapper



import collections
import math

import pytest
import torch
import torchtext.data as tt_data
from torch.utils.data import TensorDataset

from kubega.conftest import elastic_multiprocessing
from kubega.torch.data import (ElasticSampler, AdaptiveDataLoader, current_dataloader)


@pytest.mark.parametrize("num_replicas", [1, 3, 5])
@pytest.mark.parametrize("dataset_size", [9, 15, 25])
def test_sampler_epoch(num_replicas, dataset_size,
                       epoch=0, index=0, shuffle=True):
    dataset = TensorDataset(torch.rand(dataset_size))
    sampler = ElasticSampler(dataset, shuffle=shuffle)
    sampler.num_replicas = num_replicas
    sampler.set_epoch(epoch, index)
    epoch_samples = []
    sample_counts = collections.Counter()

    for rank in range(num_replicas):
        sampler.rank = rank
        epoch_samples.append(list(sampler))
        # Check indices are split evenly between replicas.
        assert len(sampler) == math.ceil((dataset_size - index % dataset_size) / num_replicas)
        # Check the actual samples obey the length.
        assert len(sampler) == len(epoch_samples[rank])
        # Check ordering is the same within the same epoch.
        assert list(sampler) == epoch_samples[rank]
        sample_counts.update(epoch_samples[rank])

    # Check all indices are present.
    assert len(sample_counts) >= dataset_size - index % dataset_size
    assert all(0 <= key < dataset_size for key in sample_counts)

    # Check each index is counted roughly the same number of times.
    assert max(sample_counts.values()) - min(sample_counts.values()) <= 1

    return epoch_samples


@pytest.mark.parametrize("num_replicas", [1, 3, 5])
@pytest.mark.parametrize("dataset_size", [9, 15, 25])
def test_sampler_shuffle(num_replicas, dataset_size):
    epoch0_samples = test_sampler_epoch(num_replicas, dataset_size, epoch=0)
    epoch1_samples = test_sampler_epoch(num_replicas, dataset_size, epoch=1)
    assert epoch0_samples != epoch1_samples  # Shuffle is on.

    epoch0_samples = test_sampler_epoch(num_replicas, dataset_size,
                                        epoch=0, shuffle=False)
    epoch1_samples = test_sampler_epoch(num_replicas, dataset_size,
                                        epoch=1, shuffle=False)
    assert epoch0_samples == epoch1_samples  # Shuffle is off.


@pytest.mark.parametrize("num_replicas", [1, 3, 5])
@pytest.mark.parametrize("dataset_size", [9, 15, 25])
def test_sampler_index(num_replicas, dataset_size):
    index = dataset_size // 2  # Set index to halfway through the dataset.
    epoch_samples = test_sampler_epoch(num_replicas, dataset_size,
                                       index=index, shuffle=False)
    samples = sum(epoch_samples, [])
    # Check contains second half of dataset.
    for idx in range(index, dataset_size):
        assert idx in samples

    index = 2 * dataset_size  # Test sampler wrap-around.
    epoch_samples = test_sampler_epoch(num_replicas, dataset_size,
                                       index=index, shuffle=False)
    assert set(sum(epoch_samples, [])) == set(range(dataset_size))


@elastic_multiprocessing
def test_dataloader_restarts():
    import kubega.checkpoint
    import kubega.collective
    from kubega.env import num_restarts, num_replicas

    kubega.collective.init("0.0.0.0")
    dataset_size = 100
    init_batch_size = 10
    dataset = TensorDataset(torch.rand(dataset_size))
    dataloader = AdaptiveDataLoader(dataset, batch_size=init_batch_size)

    # Load data samples in the following order:
    # 2 batches (20 samples) using 1 replica (local_bsz = 10, batch_size = 10)
    # 5 batches (60 samples) using 4 replica (local_bsz = 3, batch_size = 12)
    # 2 batches (20 samples) using 2 replica (local_bsz = 5, batch_size = 10)
    assert current_dataloader() is None
    idx = 0
    for idx, batch in enumerate(dataloader):
        if num_restarts() == 0 and idx == 2:
            kubega.checkpoint.save_all_states()
            return 4  # Restart with 4 replicas.
        if num_restarts() == 1 and idx == 5:
            kubega.checkpoint.save_all_states()
            return 2  # Restart with 2 replicas.
        assert current_dataloader() is dataloader._elastic

        local_bsz = batch[0].size(0)
        assert dataloader.current_local_bsz == local_bsz
        assert local_bsz == math.ceil(init_batch_size / num_replicas())
        assert dataloader.current_batch_size == num_replicas() * local_bsz
    # After the last 2 batches.
    assert idx == 1


@elastic_multiprocessing
def test_dataloader_break():
    import kubega.collective
    from kubega.env import num_restarts
    if num_restarts() == 0:
        return 2
    kubega.collective.init("0.0.0.0")
    dataset = TensorDataset(torch.rand(100))
    dataloader = AdaptiveDataLoader(dataset, batch_size=10)
    # Break in the middle of the first for-loop, and start another for-loop.
    assert current_dataloader() is None

    idx = 0
    for idx, batch in enumerate(dataloader):
        assert current_dataloader() is dataloader._elastic
        if idx == 5:
            break
    assert current_dataloader() is None
    for idx, batch in enumerate(dataloader):
        assert current_dataloader() is dataloader._elastic
    assert idx == 9  # Run 10 batches total.


@elastic_multiprocessing
def test_bptt_iterator():
    import kubega.checkpoint
    import kubega.collective
    from kubega.env import num_restarts
    from kubega.torch.adaptive_bptt_iterator import AdaptiveBPTTIterator

    kubega.collective.init("0.0.0.0")
    # Load the iterator with 500 words
    # 1 batch (5x10) using 1 replica. Restart after one iteration.
    # 9 batches (5x5) using 2 replicas to consume remaining batches.
    text_field = tt_data.Field(tokenize=tt_data.get_tokenizer("basic_english"),
                               init_token='<sos>',
                               eos_token='<eos>')
    fields = [('text', text_field)]
    examples = [tt_data.Example.fromlist([['The'] * 500], fields)]
    dataset = tt_data.Dataset(examples, fields)
    text_field.build_vocab(dataset)
    bptt_iter = AdaptiveBPTTIterator(dataset, batch_size=10, bptt_len=5)

    idx = 0
    for idx, batch in enumerate(bptt_iter):
        if num_restarts() == 0 and idx == 1:
            assert batch.text.shape == (5, 10)
            kubega.checkpoint.save_all_states()
            return 2
        if kubega.env.num_replicas() == 2:
            assert batch.text.shape == (5, 5) or batch.text.shape == (4, 5)
    if kubega.env.num_replicas() == 2:
        assert idx == 8



import os

import setuptools

"""
The python version should be 3.7 since 3.8 will cause errors in multiprocessing for macOS X.
Then, make sure you have installed
  numpy==1.21.5
  scipy==1.7.3
  pandas==1.3.5
  pytorch==1.13.1
  torchtext==0.5.0
  torchaudio==0.13.1
  torchvision==0.14.1
to avoid environment conflict.

The following packages are then required to support the *_test.py and running of this project:
  sentencepiece
  pytest
  requests
  autograd>=1.3
  portpicker>=1.3.1
  redis>=3.3.8
  semver>=2.13.0
  
The best way is to use pipreqs to generate the dependencies for kubega, and install with 
`python setup.py install --record installed.txt` or 
`pip install -r requirements.txt`.
"""


def read_requirements(file_path):
    requirements = []
    with open(file_path) as f:
        for line in f:
            if "#" in line:
                line = line[:line.index("#")]
            line = line.strip()
            if line and not line.startswith("-"):
                requirements.append(line)
    return requirements


if __name__ == '__main__':
    setuptools.setup(
        name="kubega",
        version=os.getenv("KUBEGA_VERSION", "0.0.0"),
        author="Petuum Inc. & The AdaptDL Authors",
        author_email="aurick.qiao@petuum.com",
        description="A copy of the adaptdl project "
                    "(Use is strictly limited to non-commercial areas, "
                    "only for studying and testing)",
        url="https://github.com/hliangzhao/kubega",
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: Other/Proprietary License",
            "Operating System :: POSIX :: Linux",
        ],
        packages=setuptools.find_packages(include=["kubega",
                                                   "kubega.*"]),
        python_requires='>=3.7',
        install_requires=read_requirements("requirements.txt")
    )



import math

from torchtext.data import BPTTIterator
from torchtext.data.batch import Batch
from torchtext.data.dataset import Dataset

from kubega import env
from kubega.torch.data import AdaptiveDataLoaderMixin


class AdaptiveBPTTIterator(BPTTIterator, AdaptiveDataLoaderMixin):
    """
    An adaptive DataLoader built for BPTT iterator. Used for test.
    """

    def __init__(self, dataset, batch_size, bptt_len, **kwargs):
        max_batch_size = kwargs.pop("max_batch_size", None)
        local_bsz_bounds = kwargs.pop("local_bsz_bounds", None)

        BPTTIterator.__init__(self, dataset=dataset, batch_size=batch_size,
                              bptt_len=bptt_len, **kwargs)
        AdaptiveDataLoaderMixin.__init__(self, batch_size)

        self.num_replicas = env.num_replicas()
        self.rank = env.replica_rank()

        if max_batch_size and local_bsz_bounds:
            self._elastic.autoscale_batch_size(max_batch_size,
                                               local_bsz_bounds)

    def __iter__(self):
        with self._elastic.context():
            if self._elastic.skipdone():
                return

            self.batch_size = self._elastic.sync_local_bsz()

            text = self.dataset[0].text
            text_field = self.dataset.fields['text']
            text_field.eos_token = None
            text = text + ([text_field.pad_token] *
                           int(math.ceil(len(text) / self.batch_size) *
                               self.batch_size - len(text)))
            data = text_field.numericalize(
                [text], device=self.device)
            data = data.view(self.batch_size, -1).t().contiguous()
            dataset = Dataset(examples=self.dataset.examples, fields=[
                ('text', text_field), ('target', text_field)])
            end = data.size(0)  # current length of dataset

            # Change in current batch size changes the dimensions of dataset
            # which changes the starting position in the reshaped dataset. The
            # local batch size is also a function of number of replicas, so
            # when we rescale we need to recalculate where to start the
            # iterations from for the new batch size.
            self._elastic.current_index = \
                _recompute_start(self._elastic.current_index,
                                 self._elastic.end_index, end)
            self._elastic.end_index = end

            # Every replica reads data strided by bptt_len
            start = self._elastic.current_index + (self.bptt_len * self.rank)
            step = self.bptt_len * self.num_replicas

            # The starting index of the highest rank
            highest_start = self._elastic.current_index + (self.bptt_len * (self.num_replicas - 1))

            # Number of steps we will take on the highest rank. We limit
            # iterations on all replicas by this number to prevent asymmetric
            # reduce ops which would result in a deadlock.
            min_steps_in_epoch = max(math.ceil((end - highest_start) / step), 0)  # noqa: E501
            self.iterations = 0
            while True:
                for i in range(start, end, step):
                    self.iterations += 1
                    # Make sure that _elastic.profile is called equal number of
                    # times on all replicas
                    if self.iterations > min_steps_in_epoch:
                        break
                    with self._elastic.profile(self.training and i > 0):
                        seq_len = min(self.bptt_len, data.size(0) - i - 1)
                        assert seq_len > 0
                        batch_text = data[i:i + seq_len]
                        batch_target = data[i + 1:i + 1 + seq_len]
                        if text_field.batch_first:
                            batch_text = batch_text.t().contiguous()
                            batch_target = batch_target.t().contiguous()
                        yield Batch.fromvars(
                            dataset, self.batch_size,
                            text=batch_text,
                            target=batch_target)
                        self._elastic.current_index += step

                if not self.repeat:
                    break


def _recompute_start(prev_cur, prev_end, cur_end):
    # The start index changes when there is a rescaling. We recompute a new
    # start index based on how much we covered before the restart.
    if prev_end == 0:
        return prev_cur
    return math.ceil(prev_cur * cur_end / prev_end)



import pytest

from kubega.conftest import elastic_multiprocessing


@pytest.mark.parametrize("num_replicas", [1, 2, 3, 4])
@elastic_multiprocessing
def test_profile(num_replicas):
    import kubega.checkpoint
    from kubega.env import num_restarts
    from kubega.torch._metrics import (
        profile_step_start, update_sync_time,
        profile_step_commit, _metrics_state)

    if num_restarts() == 0:
        profile = _metrics_state().profile
        assert len(profile) == 0
        # Profile local_bsz=1 but don't commit.
        profile_step_start(1)
        update_sync_time(1.0)
        # Profile local_bsz=2 and commit.
        profile_step_start(2)
        update_sync_time(1.0)
        update_sync_time(2.0)
        profile_step_commit()

        # Ensure profile is updated correctly.
        profile = _metrics_state().profile
        key = (1, 1, 2)
        assert len(profile) == 1
        assert profile[key]["accum_count"] == 0
        assert profile[key]["optim_count"] == 1
        assert profile[key]["optim_sync_time"] == 3.0
        assert profile[key]["optim_step_time"] > 0.0
        # Checkpoint and restart.
        kubega.checkpoint.save_all_states()
        return num_replicas

    elif num_restarts() == 1:
        profile = _metrics_state().profile
        # Ensure checkpoint is loaded correctly.
        key = (1, 1, 2)
        assert len(profile) == 1
        assert profile[key]["accum_count"] == 0
        assert profile[key]["optim_count"] == 1
        assert profile[key]["optim_sync_time"] == 3.0
        assert profile[key]["optim_step_time"] > 0.0

        # Profile local_bsz=3 and commit twice.
        profile_step_start(3)
        update_sync_time(2.0)
        update_sync_time(3.0)
        profile_step_commit()
        key = (1, num_replicas, 3)
        old_step_time = profile[key]["optim_step_time"]
        profile_step_start(3)
        update_sync_time(3.0)
        update_sync_time(4.0)
        profile_step_commit()

        # Ensure profile is updated correctly.
        assert len(profile) == 2
        assert profile[key]["accum_count"] == 0
        assert profile[key]["optim_count"] == 2
        assert profile[key]["optim_sync_time"] == 12.0
        assert profile[key]["optim_step_time"] > old_step_time > 0.0


@pytest.mark.parametrize("num_replicas", [1, 2, 3, 4])
@elastic_multiprocessing
def test_profile_accumulation(num_replicas):
    import kubega.checkpoint
    from kubega.env import num_restarts
    from kubega.torch._metrics import (
        profile_step_start, update_sync_time,
        profile_step_commit, _metrics_state, _fit_perf_params)

    if num_restarts() == 0:
        profile = _metrics_state().profile
        assert len(profile) == 0
        # Profile local_bsz=1 but don't commit.
        profile_step_start(1)
        update_sync_time(1.0)
        # Profile local_bsz=2 and commit.
        profile_step_start(2)
        profile_step_commit(accum_step=True)
        profile_step_start(2)
        profile_step_commit(accum_step=True)
        profile_step_start(2)
        update_sync_time(4.0)
        profile_step_commit(accum_step=False)
        profile_step_start(5)
        profile_step_commit(accum_step=True)
        profile_step_start(5)
        profile_step_commit(accum_step=True)
        profile_step_start(5)
        update_sync_time(6.0)
        profile_step_commit(accum_step=False)

        # Ensure profile is updated correctly.
        profile = _metrics_state().profile
        key = (1, 1, 2)
        assert len(profile) == 2
        assert profile[key]["accum_count"] == 2
        assert profile[key]["optim_count"] == 1
        assert profile[key]["optim_sync_time"] == 4.0
        assert profile[key]["accum_step_time"] > 0.0
        assert profile[key]["optim_step_time"] > 0.0
        profile_step_start(3)
        profile_step_commit(accum_step=True)
        profile_step_start(3)
        profile_step_commit(accum_step=True)

        # Check that fitting parameters works even
        # without a final accum_step=False commit
        for val in profile.values():
            # Ensure step time is at least sync time.
            val["optim_step_time"] += val["optim_sync_time"]
        _fit_perf_params()
        # Checkpoint and restart.
        kubega.checkpoint.save_all_states()
        return num_replicas

    elif num_restarts() == 1:
        profile = _metrics_state().profile
        # Ensure checkpoint is loaded correctly.
        key = (1, 1, 2)
        assert len(profile) == 3
        assert profile[key]["accum_count"] == 2
        assert profile[key]["optim_count"] == 1
        assert profile[key]["optim_sync_time"] == 4.0
        assert profile[key]["optim_step_time"] > 0.0

        # Profile local_bsz=3 and commit twice.
        profile_step_start(3)
        update_sync_time(2.0)
        update_sync_time(3.0)
        profile_step_commit()
        key = (1, num_replicas, 3)
        old_step_time = profile[key]["optim_step_time"]
        profile_step_start(3)
        update_sync_time(3.0)
        update_sync_time(4.0)
        profile_step_commit()

        # Ensure profile is updated correctly.
        if num_replicas == 1:
            assert len(profile) == 3
        else:
            assert len(profile) == 4
        assert profile[key]["accum_count"] == 0 if num_replicas > 1 else 2
        assert profile[key]["optim_count"] == 2
        assert profile[key]["optim_sync_time"] == 12.0
        assert profile[key]["optim_step_time"] > old_step_time > 0.0



import functools
import traceback


def print_exc(func):
    """
    A decorator that wraps the passed in function and prints any exceptions.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            traceback.print_exc()
            raise

    return wrapper
